##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Qt, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os
from distutils.version import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions
from easybuild.tools.run import run_cmd_qa
from easybuild.tools.systemtools import get_cpu_architecture, get_glibc_version, get_shared_lib_ext
from easybuild.tools.systemtools import AARCH64, POWER


class EB_Qt(ConfigureMake):
    """
    Support for building and installing Qt.
    """

    @staticmethod
    def extra_options():
        extra_vars = {
            'check_qtwebengine': [False, "Make sure QtWebEngine components is installed", CUSTOM],
            'disable_advanced_kernel_features': [False, "Disable features that require a kernel > 3.15", CUSTOM],
            'platform': [None, "Target platform to build for (e.g. linux-g++-64, linux-icc-64)", CUSTOM],
        }
        extra_vars = ConfigureMake.extra_options(extra_vars)

        # allowing to specify prefix_opt doesn't make sense for Qt, since -prefix is hardcoded in configure_step
        del extra_vars['prefix_opt']

        return extra_vars

    def configure_step(self):
        """Configure Qt using interactive `configure` script."""

        self.cfg.update('configopts', '-release')

        platform = None
        comp_fam = self.toolchain.comp_family()
        if self.cfg['platform']:
            platform = self.cfg['platform']
        # if no platform is specified, try to derive it based on compiler in toolchain
        elif comp_fam in [toolchain.GCC]:  # @UndefinedVariable
            myarch = get_cpu_architecture()
            if myarch == AARCH64:
                platform = 'linux-g++'
            else:
                platform = 'linux-g++-64'
        elif comp_fam in [toolchain.INTELCOMP]:  # @UndefinedVariable
            if LooseVersion(self.version) >= LooseVersion('4'):
                platform = 'linux-icc-64'
            else:
                platform = 'linux-icc'
                # fix -fPIC flag (-KPIC is not correct for recent Intel compilers)
                qmake_conf = os.path.join('mkspecs', platform, 'qmake.conf')
                apply_regex_substitutions(qmake_conf, [('-KPIC', '-fPIC')])

        if platform:
            self.cfg.update('configopts', "-platform %s" % platform)
        else:
            raise EasyBuildError("Don't know which platform to set based on compiler family.")

        if LooseVersion(self.version) >= LooseVersion('5.8'):
            # Qt5 doesn't respect $CFLAGS, $CXXFLAGS and $LDFLAGS, but has equivalent compiler options,
            # e.g. QMAKE_CFLAGS; see https://doc.qt.io/qt-5/qmake-variable-reference.html#qmake-cc.
            # Since EasyBuild relies e.g. for --optarch on $CFLAGS, we need to
            # set the equivalent QMAKE_* configure options.
            # (see also https://github.com/easybuilders/easybuild-easyblocks/issues/1670)
            env_to_options = {
                'CC': 'QMAKE_CC',
                'CFLAGS': 'QMAKE_CFLAGS',
                'CXX': 'QMAKE_CXX',
                'CXXFLAGS': 'QMAKE_CXXFLAGS',
                # QMAKE_LFLAGS is not a typo, see: https://doc.qt.io/qt-5/qmake-variable-reference.html#qmake-lflags
                'LDFLAGS': 'QMAKE_LFLAGS',
            }
            for env_name, option in sorted(env_to_options.items()):
                value = os.getenv(env_name)
                if value is not None:
                    if env_name.endswith('FLAGS'):
                        # For *FLAGS, we add to existing flags (e.g. those set in Qt's .pro-files).
                        config_opt = option + '+="%s"'
                    else:
                        # For compilers, we replace QMAKE_CC/CXX
                        # (otherwise, you get e.g. QMAKE_CC="g++ g++", which fails)
                        config_opt = option + '="%s"'

                    self.cfg.update('configopts', config_opt % value)

        # configure Qt such that xmlpatterns is also installed
        # -xmlpatterns is not a known configure option for Qt 5.x, but there xmlpatterns support is enabled by default
        if LooseVersion(self.version) >= LooseVersion('4') and LooseVersion(self.version) < LooseVersion('5'):
            self.cfg.update('configopts', '-xmlpatterns')

        # disable specific features to avoid that libQt5Core.so being tagged as requiring kernel 3.17,
        # which causes confusing problems like this even though the file exists and can be found by...
        #     error while loading shared libraries: libQt5Core.so.5:
        #      cannot open shared object file: No such file or directory
        # see also:
        # * https://bugs.gentoo.org/669994
        # * https://github.com/NixOS/nixpkgs/commit/a7b6a9199e8db54a798d011a0946cdeb72cfc46b
        # * https://gitweb.gentoo.org/proj/qt.git/commit/?id=9ff0752e1ee3c28818197eaaca45545708035152
        kernel_version = os.uname()[2]
        skip_kernel_features = self.cfg['disable_advanced_kernel_features']
        old_kernel_version = LooseVersion(kernel_version) < LooseVersion('3.17')
        if LooseVersion(self.version) >= LooseVersion('5.10') and (skip_kernel_features or old_kernel_version):
            self.cfg.update('configopts', '-no-feature-renameat2')
            self.cfg.update('configopts', '-no-feature-getentropy')

        cmd = "%s ./configure -prefix %s %s" % (self.cfg['preconfigopts'], self.installdir, self.cfg['configopts'])
        qa = {
            "Type 'o' if you want to use the Open Source Edition.": 'o',
            "Do you accept the terms of either license?": 'yes',
            "Which edition of Qt do you want to use?": 'o',
        }
        no_qa = [
            "for .*pro",
            r"%s.*" % os.getenv('CXX', '').replace('+', '\\+'),  # need to escape + in 'g++'
            "Reading .*",
            "WARNING .*",
            "Project MESSAGE:.*",
            "rm -f .*",
            'Creating qmake...',
            'Checking for .*...',
        ]
        run_cmd_qa(cmd, qa, no_qa=no_qa, log_all=True, simple=True, maxhits=120)

    def build_step(self):
        """Set $LD_LIBRARY_PATH before calling make, to ensure that all required libraries are found during linking."""
        # cfr. https://elist.ornl.gov/pipermail/visit-developers/2011-September/010063.html

        if LooseVersion(self.version) >= LooseVersion('5.6'):
            libdirs = ['qtbase', 'qtdeclarative']
        else:
            libdirs = ['']

        libdirs = [os.path.join(self.cfg['start_dir'], d, 'lib') for d in libdirs]
        self.cfg.update('prebuildopts', 'LD_LIBRARY_PATH=%s' % os.pathsep.join(libdirs + ['$LD_LIBRARY_PATH']))

        super(EB_Qt, self).build_step()

    def sanity_check_step(self):
        """Custom sanity check for Qt."""

        shlib_ext = get_shared_lib_ext()

        if LooseVersion(self.version) >= LooseVersion('4'):
            libversion = ''
            if LooseVersion(self.version) >= LooseVersion('5'):
                libversion = self.version.split('.')[0]

            libfile = os.path.join('lib', 'libQt%sCore.%s' % (libversion, shlib_ext))

        else:
            libfile = os.path.join('lib', 'libqt.%s' % shlib_ext)

        custom_paths = {
            'files': ['bin/moc', 'bin/qmake', libfile],
            'dirs': ['include', 'plugins'],
        }

        if self.cfg['check_qtwebengine']:
            glibc_version = get_glibc_version()
            myarch = get_cpu_architecture()
            if LooseVersion(glibc_version) <= LooseVersion("2.16"):
                self.log.debug("Skipping check for qtwebengine, since it requires a more recent glibc.")
            elif myarch == POWER:
                self.log.debug("Skipping check for qtwebengine, since it is not supported on POWER.")
            else:
                qtwebengine_libs = ['libQt%s%s.%s' % (libversion, x, shlib_ext) for x in ['WebEngine', 'WebEngineCore']]
                custom_paths['files'].extend([os.path.join('lib', lib) for lib in qtwebengine_libs])

        if LooseVersion(self.version) >= LooseVersion('4'):
            custom_paths['files'].append('bin/xmlpatterns')

        super(EB_Qt, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for DB, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError


class EB_DB(ConfigureMake):
    """Support for building and installing DB."""

    def configure_step(self):
        """Configure build: change to build dir and call configure script."""
        try:
            os.chdir('build_unix')
        except OSError as err:
            raise EasyBuildError("Failed to move to build dir: %s", err)
        super(EB_DB, self).configure_step(cmd_prefix='../dist/')



##
# Copyright 2018 Free University of Brussels
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
@author: Ward Poelmans (Free University of Brussels)
"""
import glob
import os
import stat

from easybuild.easyblocks.generic.makecp import MakeCp
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions, apply_regex_substitutions, change_dir
from easybuild.tools.filetools import copy_file, write_file
from easybuild.tools.modules import get_software_root


class EB_BWISE(MakeCp):
    """
    Custom easyblock to install BWISE
    """
    @staticmethod
    def extra_options():
        """Change default values of options"""
        extra = MakeCp.extra_options()
        # files_to_copy is not mandatory here
        extra['files_to_copy'][2] = CUSTOM
        return extra

    def build_step(self):
        """Run multiple times for different sources"""

        # BCALM is a git submodule of BWISE but we use it as a dependency because
        # it also has submodules and it's from a different developer
        bcalm = get_software_root('BCALM')
        if not bcalm:
            raise EasyBuildError("BWISE needs BCALM to work")

        makefiles_fixes = [
            (r'^CC=.*$', 'CC=$(CXX)'),
            (r'^CFLAGS=.*$', 'CFLAGS:=$(CFLAGS)'),
            (r'^LDFLAGS=.*$', 'LDFLAGS:=$(LDFLAGS) -fopenmp')
        ]

        def find_build_subdir(pattern):
            """Changes to the sub directory that matches the given pattern"""
            subdir = glob.glob(os.path.join(self.builddir, pattern))
            if subdir:
                change_dir(subdir[0])
                apply_regex_substitutions('makefile', makefiles_fixes)
                super(EB_BWISE, self).build_step()
                return subdir[0]
            else:
                raise EasyBuildError("Could not find a subdirectory matching the pattern %s", pattern)

        # BWISE has 3 independant parts, we build them one by one
        # first BWISE itself
        subdir = find_build_subdir(os.path.join('BWISE-*', 'src'))
        apply_regex_substitutions(os.path.join(subdir, '..', 'Bwise.py'),
                                  [(r'^BWISE_MAIN = .*$', 'BWISE_MAIN = os.environ[\'EBROOTBWISE\']')])

        # Onwards to BGREAT
        subdir = find_build_subdir('BGREAT2-*')
        copy_file(os.path.join(subdir, 'bgreat'), self.cfg['start_dir'])

        # Finally, BTRIM
        subdir = find_build_subdir('BTRIM-*')
        copy_file(os.path.join(subdir, 'btrim'), self.cfg['start_dir'])

        binaries = ['sequencesToNumbers', 'numbersFilter', 'path_counter', 'maximal_sr', 'simulator',
                    'path_to_kmer', 'K2000/*.py', 'K2000/*.sh']
        self.cfg['files_to_copy'] = [(['bgreat', 'btrim', 'Bwise.py'] + ['src/%s' % x for x in binaries], 'bin'),
                                     'data']

    def install_step(self):
        super(EB_BWISE, self).install_step()

        # BWISE expects BCALM to be at exactly this location...
        bcalmwrapper = """#!/bin/sh
$EBROOTBCALM/bin/bcalm "$@"
        """
        write_file(os.path.join(self.installdir, "bin", "bcalm"), bcalmwrapper)

        adjust_permissions(os.path.join(self.installdir, "bin"), stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    def sanity_check_step(self):
        """Custom sanity check for BWISE."""
        custom_paths = {
            'files': [os.path.join('bin', x) for x in ['bcalm', 'Bwise.py', 'bgreat', 'btrim']],
            'dirs': ['data']
        }
        super(EB_BWISE, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing FreeSurfer, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""

import os

from easybuild.easyblocks.generic.tarball import Tarball
from easybuild.framework.easyconfig import MANDATORY
from easybuild.tools.filetools import write_file


class EB_FreeSurfer(Tarball):
    """Support for building and installing FreeSurfer."""

    @staticmethod
    def extra_options():
        extra_vars = {
            'license_text': ['', "Text for required license file.", MANDATORY],
        }
        return Tarball.extra_options(extra_vars)

    def install_step(self):
        """Custom installation procedure for FreeSurfer, which includes installed the license file '.license'."""
        super(EB_FreeSurfer, self).install_step()
        write_file(os.path.join(self.installdir, '.license'), self.cfg['license_text'])

    def make_module_req_guess(self):
        """Include correct subdirectories to $PATH for FreeSurfer."""
        guesses = super(EB_FreeSurfer, self).make_module_req_guess()

        guesses['PATH'].extend([os.path.join('fsfast', 'bin'), os.path.join('mni', 'bin'), 'tktools'])

        return guesses

    def make_module_extra(self):
        """Define FreeSurfer-specific environment variable in generated module file."""
        txt = super(EB_FreeSurfer, self).make_module_extra()

        freesurfer_vars = {
            'FMRI_ANALYSIS_DIR': os.path.join(self.installdir, 'fsfast'),
            'FS_OVERRIDE': '0',
            'FSF_OUTPUT_FORMAT': 'nii.gz',
            'FSFAST_HOME': os.path.join(self.installdir, 'fsfast'),
            'FREESURFER_HOME': self.installdir,
            'FUNCTIONALS_DIR': os.path.join(self.installdir, 'sessions'),
            'MNI_DIR': os.path.join(self.installdir, 'mni'),
            'MNI_DATAPATH': os.path.join(self.installdir, 'mni', 'data'),
            'MINC_BIN_DIR': os.path.join(self.installdir, 'mni', 'bin'),
            'MINC_LIB_DIR': os.path.join(self.installdir, 'mni', 'lib'),
            'MNI_PERL5LIB': os.path.join(self.installdir, 'mni', 'lib', 'perl5', '5.8.5'),
            'PERL5LIB': os.path.join(self.installdir, 'mni', 'lib', 'perl5', '5.8.5'),
            'OS': 'linux',
            'SUBJECTS_DIR': os.path.join(self.installdir, 'subjects'),
        }

        for key, value in sorted(freesurfer_vars.items()):
            txt += self.module_generator.set_environment(key, freesurfer_vars[key])

        return txt

    def sanity_check_step(self):
        """Custom sanity check for FreeSurfer"""
        custom_paths = {
            'files': ['FreeSurferEnv.sh', '.license'],
            'dirs': ['bin', 'lib', 'mni'],
        }

        super(EB_FreeSurfer, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://vscentrum.be/nl/en),
# the Hercules foundation (http://www.herculesstichting.be/in_English)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for MyMediaLite, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.run import run_cmd


class EB_MyMediaLite(ConfigureMake):
    """Support for building/installing MyMediaLite."""

    def configure_step(self):
        """Custom configure step for MyMediaLite, using "make CONFIGURE_OPTIONS='...' configure"."""

        if LooseVersion(self.version) < LooseVersion('3'):
            cmd = "make CONFIGURE_OPTIONS='--prefix=%s' configure" % self.installdir
            run_cmd(cmd, log_all=True, simple=True)
        else:
            self.cfg.update('installopts', "PREFIX=%s" % self.installdir)

    def build_step(self):
        """Custom build step for MyMediaLite, using 'make all' in 'src' directory."""

        cmd = "cd src && make all && cd .."
        run_cmd(cmd, log_all=True, simple=True)

    def sanity_check_step(self):
        """Custom sanity check for MyMediaLite."""

        if LooseVersion(self.version) < LooseVersion('3'):
            bin_files = ["bin/%s_prediction" % x for x in ['item', 'mapping_item', 'mapping_rating', 'rating']]
        else:
            bin_files = ["bin/item_recommendation", "bin/rating_based_ranking", "bin/rating_prediction"]

        custom_paths = {
            'files': bin_files,
            'dirs': ["lib/mymedialite"],
        }
        super(EB_MyMediaLite, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2015-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://vscentrum.be/nl/en),
# the Hercules foundation (http://www.herculesstichting.be/in_English)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Paraver, implemented as an easyblock

@author Kenneth Hoste (Ghent University)
@author Markus Geimer (Juelich Supercomputing Centre)
@author Bernd Mohr (Juelich Supercomputing Centre)
"""
import glob
import os
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError, print_msg
from easybuild.tools.filetools import change_dir
from easybuild.tools.modules import get_software_libdir, get_software_root


class EB_Paraver(ConfigureMake):
    """Support for building/installing Paraver."""

    def run_all_steps(self, *args, **kwargs):
        """
        Put configure/build/install options in place for the 3 different components of Paraver.
        Each component lives in a separate subdirectory.
        """
        if LooseVersion(self.version) < LooseVersion('4.7'):

            # leverage support for iterated installation for older Paraver versions
            self.components = ['ptools_common_files', 'paraver-kernel', 'wxparaver']
            self.current_component = 0  # index in list above

            # initiate configopts with empty list
            self.cfg['configopts'] = []

            # first phase: build and install ptools
            # no specific configure options for the ptools component (but configopts list element must be there)
            self.cfg.update('configopts', [''])

            # second phase: build and install paraver-kernel
            self.cfg.update('configopts', ["--with-boost=%(boost)s --with-ptools-common-files=%(installdir)s"])

            # third phase: build and install wxparaver
            wxparaver_configopts = ' '.join([
                '--with-boost=%(boost)s',
                '--with-wxpropgrid=%(wxpropgrid)s',
                '--with-paraver=%(installdir)s',
            ])
            self.cfg.update('configopts', [wxparaver_configopts])
        else:
            self.components, self.current_component = None, None

        return super(EB_Paraver, self).run_all_steps(*args, **kwargs)

    def configure_step(self):
        """Custom configuration procedure for Paraver: template configuration options before using them."""

        # first, check (required) dependencies
        # this is done in configure_step to avoid breaking 'installation' with --module-only --force
        boost_root = get_software_root('Boost')
        if not boost_root:
            raise EasyBuildError("Boost is not available as a dependency")

        wx_config = None
        wxwidgets = get_software_root('wxWidgets')
        if wxwidgets:
            wx_config = os.path.join(wxwidgets, 'bin', 'wx-config')
        elif LooseVersion(self.version) >= LooseVersion('4.7'):
            raise EasyBuildError("wxWidgets is not available as a dependency")

        # determine value to pass to --with-wxpropgrid (library name)
        wxpropgrid = None
        wxpropertygrid_root = get_software_root('wxPropertyGrid')
        if wxpropertygrid_root:
            wxpropertygrid_libdir = os.path.join(wxpropertygrid_root, get_software_libdir('wxPropertyGrid'))
            libname_pattern = 'libwxcode_*_propgrid-*.so'
            wxpropgrid_libs = glob.glob(os.path.join(wxpropertygrid_libdir, libname_pattern))
            if len(wxpropgrid_libs) == 1:
                # exclude the 'lib' prefix and '.so' suffix to determine value to be passed to --with-wxpropgrid
                wxpropgrid = os.path.basename(wxpropgrid_libs[0])[3:-3]
            else:
                tup = (libname_pattern, wxpropertygrid_libdir, wxpropgrid_libs)
                raise EasyBuildError("Expected to find exactly one %s library in %s, found %s" % tup)
        else:
            self.log.info("wxPropertyGrid not included as dependency, assuming that's OK...")

        if LooseVersion(self.version) < LooseVersion('4.7'):
            component = self.components[self.current_component]
            change_dir(component)
            self.log.info("Customized start directory for component %s: %s", component, os.getcwd())

            print_msg("starting with component %s" % component, log=self.log)

            self.cfg['configopts'] = self.cfg['configopts'] % {
                'boost': boost_root,
                'installdir': self.installdir,
                'wxpropgrid': wxpropgrid,
            }
        else:
            self.cfg.update('configopts', '--with-boost=%s' % boost_root)
            self.cfg.update('configopts', '--with-paraver=%s' % self.installdir)
            self.cfg.update('configopts', '--with-wx-config=%s' % wx_config)
            # wxPropertyGrid is not required with recent wxWidgets
            if wxpropgrid:
                self.cfg.update('configopts', '--with-wxpropgrid=%s' % wxpropgrid)

        super(EB_Paraver, self).configure_step()

    def build_step(self):
        """Custom build procedure for Paraver: skip 'make' for recent versions."""

        if LooseVersion(self.version) < LooseVersion('4.7'):
            super(EB_Paraver, self).build_step()

    def install_step(self):
        """Custom installation procedure for Paraver: put symlink in place for library subdirectory."""
        super(EB_Paraver, self).install_step()

        if LooseVersion(self.version) < LooseVersion('4.7'):
            # link lib to lib64 if needed
            # this is a workaround for an issue with libtool which sometimes creates lib64 rather than lib
            if self.components[self.current_component] == self.components[0]:
                lib64dir = os.path.join(self.installdir, 'lib64')
                libdir = os.path.join(self.installdir, 'lib')

                if os.path.exists(lib64dir):
                    try:
                        self.log.debug("Symlinking %s to %s", lib64dir, libdir)
                        os.symlink(lib64dir, libdir)
                    except OSError as err:
                        raise EasyBuildError("Symlinking lib64 to lib failed: %s" % err)

            self.current_component += 1

    def sanity_check_step(self):
        """Custom sanity check for Paraver."""
        custom_paths = {
            'files': ['bin/wxparaver', 'include/paraverconfig.h', 'lib/paraver-kernel/libparaver-kernel.a'],
            'dirs': [],
        }
        super(EB_Paraver, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing WRF-Fire, implemented as an easyblock

author: Kenneth Hoste (HPC-UGent)
"""
import os

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM, MANDATORY
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, patch_perl_script_autoflush
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd, run_cmd_qa
from easybuild.tools.build_log import EasyBuildError


class EB_WRF_minus_Fire(EasyBlock):
    """Support for building/installing WRF-Fire."""

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for WRF-Fire."""
        extra_vars = {
            'buildtype': [None, "Specify the type of build (serial, smpar (OpenMP), "
                                "dmpar (MPI), dm+sm (hybrid OpenMP/MPI)).", MANDATORY],
            'runtest': [True, "Build and run WRF tests", CUSTOM],
        }
        return EasyBlock.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Add extra config options specific to WRF."""
        super(EB_WRF_minus_Fire, self).__init__(*args, **kwargs)

        self.build_in_installdir = True

    def extract_step(self):
        """Extract WRF-Fire sources."""
        self.cfg.update('unpack_options', '--strip-components=1')
        super(EB_WRF_minus_Fire, self).extract_step()

    def configure_step(self):
        """Custom configuration procedure for WRF-Fire."""

        comp_fam = self.toolchain.comp_family()

        # define $NETCDF* for netCDF dependency
        netcdf_fortran = get_software_root('netCDF-Fortran')
        if netcdf_fortran:
            env.setvar('NETCDF', netcdf_fortran)
        else:
            raise EasyBuildError("Required dependendy netCDF-Fortran is missing")

        # define $PHDF5 for parallel HDF5 dependency
        hdf5 = get_software_root('HDF5')
        if hdf5 and os.path.exists(os.path.join(hdf5, 'bin', 'h5pcc')):
            env.setvar('PHDF5', hdf5)

        # first, configure WRF part
        change_dir(os.path.join(self.cfg['start_dir'], 'WRFV3'))

        # instruct WRF-Fire to create netCDF v4 output files
        env.setvar('WRFIO_NETCDF4_FILE_SUPPORT', '1')

        # patch arch/Config_new.pl script, so that run_cmd_qa receives all output to answer questions
        patch_perl_script_autoflush(os.path.join('arch', 'Config_new.pl'))

        # determine build type option to look for
        known_build_type_options = {
            toolchain.INTELCOMP: "Linux x86_64 i486 i586 i686, ifort compiler with icc",
            toolchain.GCC: "x86_64 Linux, gfortran compiler with gcc",
            toolchain.PGI: "Linux x86_64, PGI compiler with pgcc",
        }
        build_type_option = known_build_type_options.get(comp_fam)
        if build_type_option is None:
            raise EasyBuildError("Don't know which WPS configure option to select for compiler family %s", comp_fam)

        build_type_question = r"\s*(?P<nr>[0-9]+).\s*%s\s*\(%s\)" % (build_type_option, self.cfg['buildtype'])
        qa = {
            "Compile for nesting? (1=basic, 2=preset moves, 3=vortex following) [default 1]:": '1',
        }
        std_qa = {
            # named group in match will be used to construct answer
            r"%s.*\n(.*\n)*Enter selection\s*\[[0-9]+-[0-9]+\]\s*:" % build_type_question: '%(nr)s',
        }
        run_cmd_qa('./configure', qa, std_qa=std_qa, log_all=True, simple=True)

        cpp_flag = None
        if comp_fam == toolchain.INTELCOMP:
            cpp_flag = '-fpp'
        elif comp_fam == toolchain.GCC:
            cpp_flag = '-cpp'
        else:
            raise EasyBuildError("Don't know which flag to use to specify that Fortran files were preprocessed")

        # patch configure.wrf to get things right
        comps = {
            'CFLAGS_LOCAL': os.getenv('CFLAGS'),
            'DM_FC': os.getenv('MPIF90'),
            'DM_CC': "%s -DMPI2_SUPPORT" % os.getenv('MPICC'),
            'FCOPTIM': os.getenv('FFLAGS'),
            # specify that Fortran files have been preprocessed with cpp,
            # see http://forum.wrfforum.com/viewtopic.php?f=5&t=6086
            'FORMAT_FIXED': "-FI %s" % cpp_flag,
            'FORMAT_FREE': "-FR %s" % cpp_flag,
        }
        regex_subs = [(r"^(%s\s*=\s*).*$" % k, r"\1 %s" % v) for (k, v) in comps.items()]
        apply_regex_substitutions('configure.wrf', regex_subs)

        # also configure WPS part
        change_dir(os.path.join(self.cfg['start_dir'], 'WPS'))

        # patch arch/Config_new.pl script, so that run_cmd_qa receives all output to answer questions
        patch_perl_script_autoflush(os.path.join('arch', 'Config.pl'))

        # determine build type option to look for
        known_build_type_options = {
            toolchain.INTELCOMP: "PC Linux x86_64, Intel compiler",
            toolchain.GCC: "PC Linux x86_64, g95 compiler",
            toolchain.PGI: "PC Linux x86_64 (IA64 and Opteron), PGI compiler 5.2 or higher",
        }
        build_type_option = known_build_type_options.get(comp_fam)
        if build_type_option is None:
            raise EasyBuildError("Don't know which WPS configure option to select for compiler family %s", comp_fam)

        known_wps_build_types = {
            'dmpar': 'DM parallel',
            'smpar': 'serial',
        }
        wps_build_type = known_wps_build_types.get(self.cfg['buildtype'])
        if wps_build_type is None:
            raise EasyBuildError("Don't know which WPS build type to pick for '%s'", self.cfg['builddtype'])

        build_type_question = r"\s*(?P<nr>[0-9]+).\s*%s.*%s(?!NO GRIB2)" % (build_type_option, wps_build_type)
        std_qa = {
            # named group in match will be used to construct answer
            r"%s.*\n(.*\n)*Enter selection\s*\[[0-9]+-[0-9]+\]\s*:" % build_type_question: '%(nr)s',
        }
        run_cmd_qa('./configure', {}, std_qa=std_qa, log_all=True, simple=True)

        # patch configure.wps to get things right
        comps = {
            'CC': '%s %s' % (os.getenv('MPICC'), os.getenv('CFLAGS')),
            'FC': '%s %s' % (os.getenv('MPIF90'), os.getenv('F90FLAGS'))
        }
        regex_subs = [(r"^(%s\s*=\s*).*$" % k, r"\1 %s" % v) for (k, v) in comps.items()]
        # specify that Fortran90 files have been preprocessed with cpp
        regex_subs.extend([
            (r"^(F77FLAGS\s*=\s*)", r"\1 %s " % cpp_flag),
            (r"^(FFLAGS\s*=\s*)", r"\1 %s " % cpp_flag),
        ])
        apply_regex_substitutions('configure.wps', regex_subs)

    def build_step(self):
        """Custom build procedure for WRF-Fire."""

        cmd = './compile'
        if self.cfg['parallel']:
            cmd += " -j %d" % self.cfg['parallel']

        # first, build WRF part
        change_dir(os.path.join(self.cfg['start_dir'], 'WRFV3'))
        (out, ec) = run_cmd(cmd + ' em_fire', log_all=True, simple=False, log_ok=True)

        # next, build WPS part
        change_dir(os.path.join(self.cfg['start_dir'], 'WPS'))
        (out, ec) = run_cmd('./compile', log_all=True, simple=False, log_ok=True)

    def test_step(self):
        """Custom built-in test procedure for WRF-Fire."""
        if self.cfg['runtest']:
            change_dir(os.path.join(self.cfg['start_dir'], 'WRFV3', 'test', 'em_fire', 'hill'))

            if self.cfg['buildtype'] in ['dmpar', 'smpar', 'dm+sm']:
                test_cmd = "ulimit -s unlimited && %s && %s" % (self.toolchain.mpi_cmd_for("./ideal.exe", 1),
                                                                self.toolchain.mpi_cmd_for("./wrf.exe", 2))
            else:
                test_cmd = "ulimit -s unlimited && ./ideal.exe && ./wrf.exe"
            run_cmd(test_cmd, simple=True, log_all=True, log_ok=True)

    # building/installing is done in build_step, so we can run tests
    def install_step(self):
        """Building was done in install dir, so nothing to do in install_step."""
        pass

    def sanity_check_step(self):
        """Custom sanity check for WRF-Fire."""
        custom_paths = {
            'files': [os.path.join('WRFV3', 'main', f) for f in ['ideal.exe', 'libwrflib.a', 'wrf.exe']] +
            [os.path.join('WPS', f) for f in ['geogrid.exe', 'metgrid.exe', 'ungrib.exe']],
            'dirs': [os.path.join('WRFV3', d) for d in ['main', 'run']],
        }
        super(EB_WRF_minus_Fire, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_req_guess(self):
        """Custom guesses for generated WRF-Fire module file."""
        wrf_maindir = os.path.join('WRFV3', 'main')
        return {
            'LD_LIBRARY_PATH': [wrf_maindir],
            'PATH': [wrf_maindir, 'WPS'],
        }

    def make_module_extra(self):
        """Add netCDF environment variables to module file."""
        txt = super(EB_WRF_minus_Fire, self).make_module_extra()
        netcdf_fortran = get_software_root('netCDF-Fortran')
        if netcdf_fortran:
            txt += self.module_generator.set_environment('NETCDF', netcdf_fortran)
        return txt



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing LAPACK, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""
from distutils.version import LooseVersion
import glob
import os

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.toolchains.linalg.atlas import Atlas
from easybuild.toolchains.linalg.gotoblas import GotoBLAS
from easybuild.toolchains.linalg.openblas import OpenBLAS
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import copy_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd


class EB_LAPACK(ConfigureMake):
    """
    Support for building LAPACK
    - read build_step.inc.example and replace BLAS line with configtops
    -- should be replaced by patch and variables.
    """

    @staticmethod
    def extra_options():
        extra_vars = {
            'supply_blas': [False, "Supply BLAS lib to LAPACK for building", CUSTOM],
            'test_only': [False, "Only make tests, don't try and build LAPACK lib", CUSTOM],
        }
        return ConfigureMake.extra_options(extra_vars)

    def get_blaslib(self):
        """Determine build options to define BLASLIB with"""
        blaslib = None

        for blas in [Atlas, GotoBLAS, OpenBLAS]:
            blas_modname = blas.BLAS_MODULE_NAME[0]
            blas_root = get_software_root(blas_modname)
            if blas_root:
                blas_libs = ' '.join(['-l%s' % lib for lib in blas.BLAS_LIB])
                blaslib = "-L%s %s " % (os.path.join(blas_root, 'lib'), blas_libs)
                break

        if not blaslib:
            raise EasyBuildError("No or unknown BLAS library used!")

        return blaslib

    def configure_step(self):
        """
        Configure LAPACK for build: copy build_step.inc and set make options
        """

        # copy build_step.inc file from examples
        if self.toolchain.comp_family() == toolchain.GCC:  # @UndefinedVariable
            makeinc = 'gfortran'
        elif self.toolchain.comp_family() == toolchain.INTELCOMP:  # @UndefinedVariable
            makeinc = 'ifort'
        else:
            raise EasyBuildError("Don't know which build_step.inc file to pick, unknown compiler being used...")

        src = os.path.join(self.cfg['start_dir'], 'INSTALL', 'make.inc.%s' % makeinc)
        dest = os.path.join(self.cfg['start_dir'], 'make.inc')

        if os.path.exists(dest):
            raise EasyBuildError("Destination file %s exists", dest)
        else:
            copy_file(src, dest)

        # control compiler commands being used
        self.cfg.update('buildopts', 'CC="%s"' % os.getenv('CC'))
        # older versions use FORTRAN, newer versions use FC for Fortran compiler command
        if LooseVersion(self.version) >= LooseVersion('3.9.0'):
            fc_var = 'FC'
        else:
            fc_var = 'FORTRAN'
        self.cfg.update('buildopts', '%s="%s"' % (fc_var, os.getenv('FC')))

        # set optimization flags

        self.cfg.update('buildopts', 'CFLAGS="%s"' % os.getenv('CFLAGS'))

        # Fortran compiler flags are controlled via OPTS in older version, via FFLAGS in newer versions
        if LooseVersion(self.version) >= LooseVersion('3.9.0'):
            fflags_var = 'FFLAGS'
            noopt_var = 'FFLAGS_NOOPT'
        else:
            fflags_var = 'OPTS'
            noopt_var = 'NOOPT'
        self.cfg.update('buildopts', '%s="%s"' % (fflags_var, os.getenv('FFLAGS')))

        fpic = ''
        if self.toolchain.options['pic']:
            fpic = '-fPIC'
        self.cfg.update('buildopts', '%s="%s -O0"' % (noopt_var, fpic))

        # prematurely exit configure when we're only testing
        if self.cfg['test_only']:
            self.log.info('Only testing, so skipping rest of configure.')
            return

        # supply blas lib (or not)
        if self.cfg['supply_blas']:

            blaslib = self.get_blaslib()

            self.log.debug("Providing '%s' as BLAS lib" % blaslib)
            self.cfg.update('buildopts', 'BLASLIB="%s"' % blaslib)

        else:
            self.log.debug("Not providing a BLAS lib to LAPACK.")
            self.cfg.update('buildopts', 'BLASLIB=""')

        # only build library if we're not supplying a BLAS library (otherwise testing fails)
        if not self.cfg['supply_blas']:
            self.log.info('No BLAS library provided, so only building LAPACK library (no testing).')
            self.cfg.update('buildopts', 'lib')

    # don't create a module if we're only testing
    def build_step(self):
        """
        Only build when we're not testing.
        """
        if self.cfg['test_only']:
            return

        else:
            # default make suffices (for now)
            super(EB_LAPACK, self).build_step()

    def install_step(self):
        """
        Install LAPACK: copy all .a files to lib dir in install directory
        """

        if self.cfg['test_only']:
            self.log.info('Only testing, so skipping make install.')
            pass

        srcdir = self.cfg['start_dir']
        destdir = os.path.join(self.installdir, 'lib')

        try:
            os.makedirs(destdir)

            # copy all .a files
            for libfile in glob.glob(os.path.join(srcdir, '*.a')):
                self.log.debug("Copying file %s to dir %s" % (libfile, destdir))
                copy_file(libfile, os.path.join(destdir, os.path.basename(libfile)))

            # symlink libraries to sensible names, if they aren't renamed already
            for (fromfile, tofile) in [('liblapack_LINUX.a', 'liblapack.a'),
                                       ('tmglib_LINUX.a', 'libtmglib.a')]:
                frompath = os.path.join(destdir, fromfile)
                topath = os.path.join(destdir, tofile)
                if os.path.isfile(frompath) and not os.path.isfile(tofile):
                    self.log.debug("Symlinking %s to %s" % (fromfile, tofile))
                    os.symlink(frompath, topath)

        except OSError as err:
            raise EasyBuildError("Copying %s to installation dir %s failed: %s", srcdir, destdir, err)

    def load_module(self, *args, **kwargs):
        """Don't try to load (non-existing) LAPACK module when performing a test build."""
        if not self.cfg['test_only']:
            super(EB_LAPACK, self).load_module(*args, **kwargs)

    def test_step(self):
        """
        Run BLAS and LAPACK tests that come with netlib's LAPACK.
        """
        if self.cfg['test_only']:

            if not get_software_root('LAPACK'):
                raise EasyBuildError("You need to make sure that the LAPACK module is loaded to perform testing.")

            blaslib = self.get_blaslib()

            self.log.info('Running BLAS and LAPACK tests included.')

            # run BLAS and LAPACK tests
            for lib in ["blas", "lapack"]:
                self.log.info("Running %s tests..." % lib.upper())
                cmd = "make BLASLIB='%s' %s_testing" % (blaslib, lib)
                run_cmd(cmd, log_all=True, simple=True)
        else:
            super(EB_LAPACK, self).test_step()

    # don't create a module if we're only testing
    def make_module_step(self, fake=False):
        """
        Only make LAPACK module when we're not testing.
        """
        if self.cfg['test_only']:
            pass
        else:
            return super(EB_LAPACK, self).make_module_step(fake)

    def sanity_check_step(self):
        """
        Custom sanity check for LAPACK (only run when not testing)
        """
        if not self.cfg['test_only']:
            custom_paths = {
                'files': ["lib/%s" % x for x in ["liblapack.a", "libtmglib.a"]],
                'dirs': []
            }

            super(EB_LAPACK, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2015-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing pbdMPI, implemented as an easyblock

@author: Ewan Higgs (Ghent University)
@author: Peter Maxwell (University of Auckland)
"""

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.rpackage import RPackage


class EB_pbdMPI(RPackage):
    """Support for building/installing pbdMPI."""

    def configure_step(self):
        """Configure Step of build process for pbdMPI."""
        mpi_types = {
            toolchain.INTELMPI: 'INTELMPI',
            toolchain.MPI_TYPE_MPICH: 'MPICH',
            toolchain.MPI_TYPE_OPENMPI: 'OPENMPI',
        }
        mpi_type = mpi_types[self.toolchain.mpi_family()]
        self.configureargs.extend([
            "--with-mpi-include=%s" % self.toolchain.get_variable('MPI_INC_DIR'),
            "--with-mpi-libpath=%s" % self.toolchain.get_variable('MPI_LIB_DIR'),
            "--with-mpi=%s" % self.toolchain.get_software_root(self.toolchain.MPI_MODULE_NAME)[0],
            "--with-mpi-type=%s" % mpi_type,
        ])

        super(EB_pbdMPI, self).configure_step()

    def run(self):
        """Configure before installing pbdMPI as an extension."""
        self.configure_step()
        super(EB_pbdMPI, self).run()



##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Bowtie, implemented as an easyblock

@author: Cedric Laczny (Uni.Lu)
@author: Fotis Georgatos (Uni.Lu)
@author: Kenneth Hoste (Ghent University)
@author: Jens Timmerman (Ghent University)
"""
from distutils.version import LooseVersion
import glob
import os
import shutil

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import mkdir


class EB_Bowtie(ConfigureMake):
    """
    Support for building bowtie (ifast and sensitive read alignment)
    """

    def configure_step(self):
        """
        Set compilers in buildopts, there is no configure script.
        """
        comp_opts = 'CC="%(cc)s" CXX="%(cxx)s" CPP="%(cxx)s"' % {'cc': os.getenv('CC'), 'cxx': os.getenv('CXX')}
        self.cfg.update('buildopts', comp_opts)

        # make sure install target is specified for recent Bowtie versions that support 'make install'
        if LooseVersion(self.version) >= LooseVersion('1.1.2'):
            self.cfg.update('installopts', "prefix=%s" % self.installdir)

    def install_step(self):
        """
        Install by copying files to install dir
        """
        if LooseVersion(self.version) >= LooseVersion('1.1.2'):
            # 'make install' is supported since Bowtie 1.1.2
            super(EB_Bowtie, self).install_step()
        else:
            destdir = os.path.join(self.installdir, 'bin')
            mkdir(destdir)
            try:
                glob_pat = os.path.join(self.cfg['start_dir'], 'bowtie*')
                binaries = [x for x in glob.glob(glob_pat) if os.path.splitext(x)[0] == x]
                self.log.debug("Copying binaries to %s: %s", destdir, binaries)
                for binary in binaries:
                    shutil.copy2(binary, destdir)

            except (IOError, OSError) as err:
                raise EasyBuildError("Copying binaries to installation dir %s failed: %s", destdir, err)

    def sanity_check_step(self):
        """Custom sanity check for Bowtie."""
        binaries = ['bowtie', 'bowtie-build', 'bowtie-inspect']
        if LooseVersion(self.version) > LooseVersion('1.1.0'):
            binaries.extend(['bowtie-align-l', 'bowtie-align-s', 'bowtie-build-l', 'bowtie-build-s',
                             'bowtie-inspect-l', 'bowtie-inspect-s'])
        custom_paths = {
            'files': [os.path.join('bin', x) for x in binaries],
            'dirs': []
        }
        super(EB_Bowtie, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2013 the Cyprus Institute
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
@author: George Tsouloupas (The Cyprus Institute)
@author: Fotis Georgatos (Uni.Lu, NTUA)
@author: Kenneth Hoste (Ghent University)
"""
import os
import glob

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import BUILD, MANDATORY
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import change_dir, copy_dir, copy_file, mkdir
from easybuild.tools.py2vs3 import string_type


class MakeCp(ConfigureMake):
    """
    Software with no configure and no make install step.
    """
    @staticmethod
    def extra_options(extra_vars=None):
        """
        Define list of files or directories to be copied after make
        """
        extra = {
            'files_to_copy': [None, "List of files or dirs to copy", MANDATORY],
            'with_configure': [False, "Run configure script before building", BUILD],
        }
        if extra_vars is None:
            extra_vars = {}
        extra.update(extra_vars)
        return ConfigureMake.extra_options(extra_vars=extra)

    def configure_step(self, cmd_prefix=''):
        """
        Configure build if required
        """
        if self.cfg.get('with_configure', False):
            return super(MakeCp, self).configure_step(cmd_prefix=cmd_prefix)

    def install_step(self):
        """Install by copying specified files and directories."""

        # make sure we're (still) in the start dir
        change_dir(self.cfg['start_dir'])

        files_to_copy = self.cfg.get('files_to_copy') or []
        self.log.debug("Starting install_step with files_to_copy: %s", files_to_copy)
        for fil in files_to_copy:
            if isinstance(fil, tuple):
                # ([src1, src2], targetdir)
                if len(fil) == 2 and isinstance(fil[0], list) and isinstance(fil[1], string_type):
                    files_specs = fil[0]
                    target = os.path.join(self.installdir, fil[1])
                else:
                    raise EasyBuildError("Only tuples of format '([<source files>], <target dir>)' supported.")
            # 'src_file' or 'src_dir'
            elif isinstance(fil, string_type):
                files_specs = [fil]
                target = self.installdir
            else:
                raise EasyBuildError("Found neither string nor tuple as file to copy: '%s' (type %s)", fil, type(fil))

            mkdir(target, parents=True)

            for orig_files_spec in files_specs:
                if isinstance(orig_files_spec, tuple):
                    files_spec = orig_files_spec[0]
                    dest = orig_files_spec[1]
                else:
                    files_spec = orig_files_spec
                    dest = None

                # first look for files in start dir
                filepaths = glob.glob(os.path.join(self.cfg['start_dir'], files_spec))
                tup = (files_spec, self.cfg['start_dir'], filepaths)
                self.log.debug("List of files matching '%s' in start dir %s: %s" % tup)

                if not filepaths and len(self.src) > 0 and 'finalpath' in self.src[0]:
                    # use location of first unpacked source file as fallback location
                    tup = (files_spec, self.cfg['start_dir'])
                    self.log.warning("No files matching '%s' found in start dir %s" % tup)
                    filepaths = glob.glob(os.path.join(self.src[0]['finalpath'], files_spec))
                    self.log.debug("List of files matching '%s' in %s: %s" % (tup + (filepaths,)))

                # there should be at least one match per file spec
                if not filepaths:
                    raise EasyBuildError("No files matching '%s' found anywhere.", files_spec)

                if dest and len(filepaths) != 1:
                    raise EasyBuildError("When a list with new names has been specified, the original file spec can "
                                         "only match a single file yet it gives: %s", filepaths)

                for filepath in filepaths:
                    # copy individual file
                    if os.path.isfile(filepath):
                        if dest:
                            target_dest = os.path.join(target, dest)
                        else:
                            target_dest = target
                        self.log.debug("Copying file %s to %s", filepath, target_dest)
                        copy_file(filepath, target_dest)
                    # copy directory
                    elif os.path.isdir(filepath):
                        self.log.debug("Copying directory %s to %s", filepath, target)
                        fulltarget = os.path.join(target, os.path.basename(filepath))
                        copy_dir(filepath, fulltarget, symlinks=self.cfg['keepsymlinks'])
                    else:
                        raise EasyBuildError("Can't copy non-existing path %s to %s", filepath, target)



##
# Copyright 2014-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of the University of Ghent (http://ugent.be/hpc).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
#
# This work implements a part of the HPCBIOS project and is a component of the policy:
# http://hpcbios.readthedocs.org/en/latest/HPCBIOS_2012-94.html
##
"""
EasyBuild support for installing Modeller, implemented as an easyblock

@author: Pablo Escobar Lopez (SIB - University of Basel)
"""

import os

from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd_qa


class EB_Modeller(EasyBlock):
    """Support for installing Modeller."""

    def configure_step(self):
        """ Skip configuration step """
        pass

    def build_step(self):
        """ Skip build step """
        pass

    def install_step(self):
        """Interactive install of Modeller."""

        if self.cfg['key'] is None:
            raise EasyBuildError("Easyconfig parameter 'key' is not defined")

        cmd = "%s/Install" % self.cfg['start_dir']

        # by default modeller tries to install to $HOME/bin/modeller9.13
        # get this path to use it in the question/answer
        default_install_path = os.path.join(os.path.expanduser('~'), 'bin', 'modeller%s' % self.cfg['version'])

        qa = {
            # installer will autodetect the right arch. [3] = x86_64
            'Select the type of your computer from the list above [3]:': '',
            "[%s]:" % default_install_path: self.installdir,
            'http://salilab.org/modeller/registration.html:': self.cfg["key"],
            'https://salilab.org/modeller/registration.html:': self.cfg["key"],
            'Press <Enter> to begin the installation:': '',
            'Press <Enter> to continue:': ''
        }

        run_cmd_qa(cmd, qa, log_all=True, simple=True)

    def sanity_check_step(self):
        """Custom sanity check for Modeller."""
        custom_paths = {
            'files': ["bin/mod%s" % self.version, "bin/modpy.sh", "bin/modslave.py"],
            'dirs': ["doc", "lib", "examples"],
        }
        super(EB_Modeller, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_req_guess(self):
        """Custom guesses for environment variables (PYTHONPATH, LD_LIBRARY_PATH) for modeller."""
        guesses = super(EB_Modeller, self).make_module_req_guess()

        libspath = os.path.join(self.installdir, 'lib')
        if os.path.exists(libspath):
            libdirs = os.listdir(libspath)
            if len(libdirs) == 1:
                libdir = libdirs[0]
            else:
                raise EasyBuildError("Failed to isolate a single libdir from list of 'lib' subdirectories: %s", libdirs)

            py2libdirs = [d for d in os.listdir(os.path.join(libspath, libdir)) if d.startswith('python2')]
            if len(py2libdirs) >= 1:
                py2libdir = py2libdirs[-1]
            else:
                raise EasyBuildError("Failed to isolate latest Python lib dir from list %s", py2libdirs)

            guesses.update({
                'PYTHONPATH': [os.path.join('lib', libdir, py2libdir), "modlib"],
                'LD_LIBRARY_PATH': [os.path.join('lib', libdir)],
            })

        return guesses



##
# Copyright 2021 Vrije Universiteit Brussel
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing a wrapper module file for OpenSSL

@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os
import re

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.filetools import change_dir, expand_glob_paths, mkdir, read_file, symlink, which
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import DARWIN, LINUX, get_os_type, get_shared_lib_ext, find_library_path
from easybuild.tools.build_log import EasyBuildError, print_warning


class EB_OpenSSL_wrapper(Bundle):
    """
    Locate the installation files of OpenSSL in the host system
    If available, wrap the system OpenSSL by symlinking all installation files
    Fall back to the bundled component otherwise
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to OpenSSL wrapper"""
        extra_vars = Bundle.extra_options(extra_vars=extra_vars)
        extra_vars.update({
            'wrap_system_openssl': [True, 'Detect and wrap OpenSSL installation in host system', CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Locate the installation files of OpenSSL in the host system"""
        super(EB_OpenSSL_wrapper, self).__init__(*args, **kwargs)

        # Libraries packaged in OpenSSL
        openssl_libs = ['libssl', 'libcrypto']
        # list of relevant library extensions per system and version of OpenSSL
        # the first item should be the extension of an installation from source,
        # it will be used in the sanity checks of the component
        openssl_libext = {
            '1.0': {
                LINUX: ('so.1.0.0', 'so.10'),
                DARWIN: ('1.0.dylib', ),
            },
            '1.1': {
                LINUX: ('so.1.1', ),
                DARWIN: ('1.1.dylib', ),
            },
        }

        os_type = get_os_type()
        if self.version in openssl_libext and os_type in openssl_libext[self.version]:
            # generate matrix of versioned .so filenames
            system_versioned_libs = [
                ['%s.%s' % (lib, ext) for ext in openssl_libext[self.version][os_type]]
                for lib in openssl_libs
            ]
            self.log.info("Matrix of version library names: %s", system_versioned_libs)
        else:
            raise EasyBuildError("Don't know name of OpenSSL system library for version %s and OS type %s",
                                 self.version, os_type)

        # by default target the first option of each OpenSSL library,
        # which corresponds to installation from source
        self.target_ssl_libs = [lib_name[0] for lib_name in system_versioned_libs]
        self.log.info("Target OpenSSL libraries: %s", self.target_ssl_libs)

        # folders containing engines libraries
        openssl_engines = {
            '1.0': 'engines',
            '1.1': 'engines-1.1',
        }
        self.target_ssl_engine = openssl_engines[self.version]

        # Paths to system libraries and headers of OpenSSL
        self.system_ssl = {
            'bin': None,
            'engines': None,
            'include': None,
            'lib': None,
        }

        # early return when we're not wrapping the system OpenSSL installation
        if not self.cfg.get('wrap_system_openssl'):
            self.log.info("Not wrapping system OpenSSL installation!")
            return

        # Check the system libraries of OpenSSL
        # (only needs first library in openssl_libs to find path to libraries)
        for idx, libssl in enumerate(system_versioned_libs[0]):
            self.system_ssl['lib'] = find_library_path(libssl)
            if self.system_ssl['lib']:
                # change target libraries to the ones found in the system
                self.target_ssl_libs = [lib_name[idx] for lib_name in system_versioned_libs]
                self.log.info("Target system OpenSSL libraries: %s", self.target_ssl_libs)
                break

        if self.system_ssl['lib']:
            self.log.info("Found library '%s' in: %s", openssl_libs[0], self.system_ssl['lib'])
        else:
            self.log.info("OpenSSL library '%s' not found, falling back to OpenSSL in EasyBuild", openssl_libs[0])

        # Directory with engine libraries
        if self.system_ssl['lib']:
            lib_dir = os.path.dirname(self.system_ssl['lib'])
            lib_engines_dir = [
                os.path.join(lib_dir, 'openssl', self.target_ssl_engine),
                os.path.join(lib_dir, self.target_ssl_engine),
            ]

            for engines_path in lib_engines_dir:
                if os.path.isdir(engines_path):
                    self.system_ssl['engines'] = engines_path
                    self.log.debug("Found OpenSSL engines in: %s", self.system_ssl['engines'])
                    break

            if not self.system_ssl['engines']:
                self.system_ssl['lib'] = None
                self.log.info("OpenSSL engines not found in host system, falling back to OpenSSL in EasyBuild")

        # Check system include paths for OpenSSL headers
        cmd = "LC_ALL=C gcc -E -Wp,-v -xc /dev/null"
        (out, ec) = run_cmd(cmd, log_all=True, simple=False, trace=False)

        sys_include_dirs = []
        for match in re.finditer(r'^\s(/[^\0\n]*)+', out, re.MULTILINE):
            sys_include_dirs.extend(match.groups())
        self.log.debug("Found the following include directories in host system: %s", ', '.join(sys_include_dirs))

        # headers are located in 'include/openssl' by default
        ssl_include_subdirs = [self.name.lower()]
        if self.version == '1.1':
            # but version 1.1 can be installed in 'include/openssl11/openssl' as well, for example in CentOS 7
            # prefer 'include/openssl' as long as the version of headers matches
            ssl_include_subdirs.append(os.path.join('openssl11', self.name.lower()))

        ssl_include_dirs = [os.path.join(incd, subd) for incd in sys_include_dirs for subd in ssl_include_subdirs]
        ssl_include_dirs = [include for include in ssl_include_dirs if os.path.isdir(include)]

        # find location of header files, verify that the headers match our OpenSSL version
        openssl_version_regex = re.compile(r"SHLIB_VERSION_NUMBER\s\"([0-9]+\.[0-9]+)", re.M)
        for include_dir in ssl_include_dirs:
            opensslv_path = os.path.join(include_dir, 'opensslv.h')
            self.log.debug("Checking OpenSSL version in %s...", opensslv_path)
            if os.path.exists(opensslv_path):
                opensslv = read_file(opensslv_path)
                header_majmin_version = openssl_version_regex.search(opensslv)
                if header_majmin_version:
                    header_majmin_version = header_majmin_version.group(1)
                    if re.match('^' + header_majmin_version, self.version):
                        self.system_ssl['include'] = include_dir
                        self.log.info("Found OpenSSL headers in host system: %s", self.system_ssl['include'])
                        break
                    else:
                        self.log.debug("Header major/minor version '%s' doesn't match with %s",
                                       header_majmin_version, self.version)
                else:
                    self.log.debug("Pattern '%s' not found in %s", openssl_version_regex.pattern, opensslv_path)
            else:
                self.log.info("OpenSSL header file %s not found", opensslv_path)

        if not self.system_ssl['include']:
            self.log.info("OpenSSL headers not found in host system, falling back to OpenSSL in EasyBuild")

        # Check system OpenSSL binary
        if self.version == '1.1':
            # prefer 'openssl11' over 'openssl' with v1.1
            self.system_ssl['bin'] = which('openssl11')

        if not self.system_ssl['bin']:
            self.system_ssl['bin'] = which(self.name.lower())

        if self.system_ssl['bin']:
            self.log.info("System OpenSSL binary found: %s", self.system_ssl['bin'])
        else:
            self.log.info("System OpenSSL binary not found!")

    def fetch_step(self, *args, **kwargs):
        """Fetch sources if OpenSSL component is needed"""
        if not all([self.system_ssl['lib'], self.system_ssl['include']]):
            super(EB_OpenSSL_wrapper, self).fetch_step(*args, **kwargs)

    def extract_step(self):
        """Extract sources if OpenSSL component is needed"""
        if not all([self.system_ssl['lib'], self.system_ssl['include']]):
            super(EB_OpenSSL_wrapper, self).extract_step()

    def install_step(self):
        """Symlink target OpenSSL installation"""

        if all(self.system_ssl[key] for key in ('bin', 'engines', 'include', 'lib')):
            # note: symlink to individual files, not directories,
            # since directory symlinks get resolved easily...

            # link OpenSSL libraries in system
            lib64_dir = os.path.join(self.installdir, 'lib64')
            lib64_engines_dir = os.path.join(lib64_dir, os.path.basename(self.system_ssl['engines']))
            mkdir(lib64_engines_dir, parents=True)

            # link existing known libraries
            ssl_syslibdir = os.path.dirname(self.system_ssl['lib'])
            lib_files = [os.path.join(ssl_syslibdir, x) for x in self.target_ssl_libs]
            for libso in lib_files:
                symlink(libso, os.path.join(lib64_dir, os.path.basename(libso)))

            # link engines library files
            engine_lib_pattern = [os.path.join(self.system_ssl['engines'], '*')]
            for engine_lib in expand_glob_paths(engine_lib_pattern):
                symlink(engine_lib, os.path.join(lib64_engines_dir, os.path.basename(engine_lib)))

            # relative symlink for unversioned libraries
            cwd = change_dir(lib64_dir)
            for libso in self.target_ssl_libs:
                unversioned_lib = '%s.%s' % (libso.split('.')[0], get_shared_lib_ext())
                symlink(libso, unversioned_lib, use_abspath_source=False)
            change_dir(cwd)

            # link OpenSSL headers in system
            include_dir = os.path.join(self.installdir, 'include', self.name.lower())
            mkdir(include_dir, parents=True)
            include_pattern = [os.path.join(self.system_ssl['include'], '*')]
            for header_file in expand_glob_paths(include_pattern):
                symlink(header_file, os.path.join(include_dir, os.path.basename(header_file)))

            # link OpenSSL binary in system
            bin_dir = os.path.join(self.installdir, 'bin')
            mkdir(bin_dir)
            symlink(self.system_ssl['bin'], os.path.join(bin_dir, self.name.lower()))

        else:
            # install OpenSSL component
            print_warning("Not all OpenSSL components found, falling back to OpenSSL in EasyBuild!")
            super(EB_OpenSSL_wrapper, self).install_step()

    def sanity_check_step(self):
        """Custom sanity check for OpenSSL wrapper."""
        shlib_ext = get_shared_lib_ext()

        ssl_libs = ['%s.%s' % (libso.split('.')[0], shlib_ext) for libso in self.target_ssl_libs]
        ssl_libs.extend(self.target_ssl_libs)

        ssl_files = [os.path.join('bin', self.name.lower())]
        ssl_files.extend(os.path.join('lib', libso) for libso in ssl_libs)

        ssl_dirs = [
            os.path.join('include', self.name.lower()),
            os.path.join('lib', self.target_ssl_engine),
        ]

        custom_paths = {
            'files': ssl_files,
            'dirs': ssl_dirs,
        }

        # make sure that version mentioned in output of 'openssl version' matches version we are using
        custom_commands = ["ssl_ver=$(openssl version); [ ${ssl_ver:8:3} == '%s' ]" % self.version[:3]]

        super(Bundle, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2015-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing SCons, implemented as an easyblock

@author: Balazs Hajgato (Free University Brussels (VUB))
"""
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.run import run_cmd


class SCons(EasyBlock):
    """Support for building/installing with SCons."""

    @staticmethod
    def extra_options():
        extra_vars = {
            'prefix_arg': ['PREFIX=', "Syntax for specifying installation prefix", CUSTOM],
        }
        return EasyBlock.extra_options(extra_vars)

    def configure_step(self):
        """
        No configure step for SCons
        """
        if self.cfg['prefix_arg']:
            self.prefix = self.cfg['prefix_arg'] + self.installdir
        else:
            self.prefix = ''

    def build_step(self, verbose=False):
        """
        Build with SCons
        """

        par = ''
        if self.cfg['parallel']:
            par = "-j %s" % self.cfg['parallel']

        cmd = "%(prebuildopts)s scons %(par)s %(buildopts)s %(prefix)s" % {
            'buildopts': self.cfg['buildopts'],
            'prebuildopts': self.cfg['prebuildopts'],
            'prefix': self.prefix,
            'par': par,
        }
        (out, _) = run_cmd(cmd, log_all=True, log_output=verbose)

        return out

    def test_step(self):
        """
        Test with SCons
        """
        if self.cfg['runtest']:
            cmd = "%s scons %s %s" % (self.cfg['pretestopts'], self.cfg['runtest'], self.cfg['testopts'])
            run_cmd(cmd, log_all=True)

    def install_step(self):
        """
        Install with SCons
        """
        cmd = "%(preinstallopts)s scons %(prefix)s install %(installopts)s" % {
            'installopts': self.cfg['installopts'],
            'preinstallopts': self.cfg['preinstallopts'],
            'prefix': self.prefix,
        }
        (out, _) = run_cmd(cmd, log_all=True)

        return out



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing COMSOL, implemented as an easyblock

@author: Mikael OEhman (Chalmers University of Technology)
@author: Ake Sandgren (HPC2N, Umea University)
"""
import os
import re
import stat

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.packedbinary import PackedBinary
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions, copy_file, find_flexlm_license, read_file, write_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_COMSOL(PackedBinary):
    """Support for installing COMSOL."""

    def __init__(self, *args, **kwargs):
        """Add extra config options specific to COMSOL."""
        super(EB_COMSOL, self).__init__(*args, **kwargs)
        self.configfile = os.path.join(self.builddir, 'my_setupconfig.ini')

    def configure_step(self):
        """Configure COMSOL installation: create license file."""

        # The tar file comes from the DVD and has 0444 as permission at the top dir.
        adjust_permissions(self.start_dir, stat.S_IWUSR)

        default_lic_env_var = 'LMCOMSOL_LICENSE_FILE'
        lic_specs, self.license_env_var = find_flexlm_license(custom_env_vars=[default_lic_env_var],
                                                              lic_specs=[self.cfg['license_file']])

        if lic_specs:
            if self.license_env_var is None:
                self.log.info("Using COMSOL license specifications from 'license_file': %s", lic_specs)
                self.license_env_var = default_lic_env_var
            else:
                self.log.info("Using COMSOL license specifications from $%s: %s", self.license_env_var, lic_specs)

            self.license_file = os.pathsep.join(lic_specs)
            env.setvar(self.license_env_var, self.license_file)
        else:
            msg = "No viable license specifications found; "
            msg += "specify 'license_file', or define $%s" % default_lic_env_var
            raise EasyBuildError(msg)

        copy_file(os.path.join(self.start_dir, 'setupconfig.ini'), self.configfile)
        config = read_file(self.configfile)

        config_vars = {
            'agree': '1',
            'desktopshortcuts': '0',
            'fileassoc': '0',
            'firewall': '0',
            'installdir': self.installdir,
            'license': self.license_file,
            'licmanager': '0',
            'linuxlauncher': '0',
            'showgui': '0',
            'startmenushortcuts': '0',
            'symlinks': '0',
        }

        matlab_root = get_software_root("MATLAB")
        if matlab_root:
            config_vars.update({'matlabdir': matlab_root})

        for key, val in config_vars.items():
            regex = re.compile(r"^%s\s*=.*" % key, re.M)
            config = regex.sub("%s=%s" % (key, val), config)

        write_file(self.configfile, config)

        self.log.debug('configuration file written to %s:\n %s', self.configfile, config)

    def install_step(self):
        """COMSOL install procedure using 'install' command."""

        setup_script = os.path.join(self.start_dir, 'setup')

        # make sure setup script is executable
        adjust_permissions(setup_script, stat.S_IXUSR)

        # make sure $DISPLAY is not defined, which may lead to (hard to trace) problems
        # this is a workaround for not being able to specify --nodisplay to the install scripts
        env.unset_env_vars(['DISPLAY'])

        cmd = ' '.join([self.cfg['preinstallopts'], setup_script, '-s', self.configfile, self.cfg['installopts']])
        run_cmd(cmd, log_all=True, simple=True)

    def sanity_check_step(self):
        """Custom sanity check for COMSOL."""
        custom_paths = {
            'files': [
                "bin/comsol", "bin/glnxa64/comsol",
                "lib/glnxa64/libcscomsolgeom.%s" % get_shared_lib_ext(),
            ],
            'dirs': ["java/glnxa64", "plugins"],
        }
        super(EB_COMSOL, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing ANSYS, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Bart Verleye (Centre for eResearch, Auckland)
"""
import os
import re
import stat
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.packedbinary import PackedBinary
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions
from easybuild.tools.run import run_cmd


class EB_ANSYS(PackedBinary):
    """Support for installing ANSYS."""

    def __init__(self, *args, **kwargs):
        """Initialize ANSYS-specific variables."""
        super(EB_ANSYS, self).__init__(*args, **kwargs)
        self.ansysver = None

    def install_step(self):
        """Custom install procedure for ANSYS."""
        licserv = self.cfg['license_server']
        if licserv is None:
            licserv = os.getenv('EB_ANSYS_LICENSE_SERVER', 'license.example.com')
        licport = self.cfg['license_server_port']
        if licport is None:
            licport = os.getenv('EB_ANSYS_LICENSE_SERVER_PORT', '2325:1055')

        cmd = "./INSTALL -silent -install_dir %s -licserverinfo %s:%s" % (self.installdir, licport, licserv)
        run_cmd(cmd, log_all=True, simple=True)

        adjust_permissions(self.installdir, stat.S_IWOTH, add=False)

    def set_ansysver(self):
        """Determine name of version subdirectory in installation directory."""
        if LooseVersion(self.version) >= LooseVersion('2019'):
            entries = os.listdir(self.installdir)
            version_dir_regex = re.compile('^v[0-9]+')
            version_dirs = [e for e in entries if version_dir_regex.match(e)]
            if len(version_dirs) == 1:
                self.ansysver = version_dirs[0]
            else:
                raise EasyBuildError("Failed to isolate version subdirectory in %s: %s", self.installdir, entries)
        else:
            self.ansysver = 'v' + ''.join(self.version.split('.')[0:2])

    def make_module_req_guess(self):
        """Custom extra module file entries for ANSYS."""

        if self.ansysver is None:
            self.set_ansysver()

        guesses = super(EB_ANSYS, self).make_module_req_guess()
        dirs = [
            'tgrid/bin',
            'Framework/bin/Linux64',
            'aisol/bin/linx64',
            'RSM/bin',
            'ansys/bin',
            'autodyn/bin',
            'CFD-Post/bin',
            'CFX/bin',
            'fluent/bin',
            'TurboGrid/bin',
            'polyflow/bin',
            'Icepak/bin',
            'icemcfd/linux64_amd/bin'
        ]
        if LooseVersion(self.version) >= LooseVersion('19.0'):
            dirs.append('CEI/bin')

        guesses['PATH'] = [os.path.join(self.ansysver, d) for d in dirs]

        return guesses

    def make_module_extra(self):
        """Define extra environment variables required by Ansys"""

        if self.ansysver is None:
            self.set_ansysver()

        txt = super(EB_ANSYS, self).make_module_extra()
        icem_acn = os.path.join(self.installdir, self.ansysver, 'icemcfd', 'linux64_amd')
        txt += self.module_generator.set_environment('ICEM_ACN', icem_acn)
        return txt

    def sanity_check_step(self):
        """Custom sanity check for ANSYS."""

        if self.ansysver is None:
            self.set_ansysver()

        custom_paths = {
            'files': [os.path.join(self.ansysver, 'fluent', 'bin', 'fluent%s' % x) for x in ['', '_arch', '_sysinfo']],
            'dirs': [os.path.join(self.ansysver, x) for x in ['ansys', 'aisol', 'CFD-Post', 'CFX']]
        }
        super(EB_ANSYS, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing CHARMM, implemented as an easyblock

@author: Ward Poelmans (Ghent University)
"""
# TODO: add support for more QC software (q-chem, gamess, ...)

import shutil

from easybuild.framework.easyconfig import CUSTOM
from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd
import easybuild.tools.toolchain as toolchain

# Possible systemsizes for CHARMM
KNOWN_SYSTEM_SIZES = ['huge', 'xxlarge', 'xlarge', 'large', 'medium', 'small', 'xsmall', 'reduce']


class EB_CHARMM(EasyBlock):
    """
    Support for building/installing CHARMM
    """

    @staticmethod
    def extra_options():
        """Add extra easyconfig parameters custom to CHARMM."""
        extra_vars = {
            'build_options': ["FULL", "Specify the options to the build script", CUSTOM],
            'system_size': ["medium", "Specify the supported systemsize: %s" % ', '.join(KNOWN_SYSTEM_SIZES), CUSTOM],
            'runtest': [True, "Run tests after each build", CUSTOM],
        }
        return EasyBlock.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for CHARMM."""
        super(EB_CHARMM, self).__init__(*args, **kwargs)
        self.arch = 'UNKNOWN'

    def configure_step(self):
        # Clean out old dir but don't create new one
        self.cfg['dontcreateinstalldir'] = True

        if self.toolchain.comp_family() == toolchain.INTELCOMP:
            self.arch = "em64t"
        else:
            self.arch = "gnu"

        super(EB_CHARMM, self).make_dir(self.installdir, True, dontcreateinstalldir=True)

    def build_step(self, verbose=False):
        """Start the actual build"""
        if self.cfg['system_size'] not in KNOWN_SYSTEM_SIZES:
            raise EasyBuildError("Unknown system size '%s' specified, known: %s",
                                 self.cfg['system_size'], KNOWN_SYSTEM_SIZES)

        self.log.info("Building for size: %s" % self.cfg['system_size'])
        self.log.info("Build options from the easyconfig: %s" % self.cfg['build_options'])
        build_options = self.cfg['build_options']

        # FFTW and MKL are mutally exclusive
        if get_software_root("FFTW"):
            self.log.info("Using FFTW")
            build_options += " FFTW"
        else:
            self.log.info("Not using FFTW")
            if get_software_root("imkl"):
                self.log.info("Using the MKL")
                build_options += " MKL"
            else:
                self.log.info("Not using MKL")

        # Currently, only support for g09 added
        if get_software_root("Gaussian") and 'g09' in get_software_version('Gaussian'):
            self.log.info("Using g09")
            build_options += " G09"
        else:
            self.log.info("Not using g09")

        if self.toolchain.options.get('usempi', None):
            self.log.info("Using MPI")
            # M means use MPI and MPIF90 means let mpif90 handle all MPI stuff
            build_options += " M MPIF90"

        # By default, CHARMM uses gfortran. We need to specify if we want ifort
        if self.toolchain.comp_family() == toolchain.INTELCOMP:
            build_options += " IFORT"

        cmd = ' '.join([self.cfg['prebuildopts'], './install.com', self.arch, self.cfg['system_size'], build_options])
        (out, _) = run_cmd(cmd, log_all=True, simple=False, log_output=verbose)
        return out

    def test_step(self):
        """Run the testsuite"""
        # Allow to skip the tests by setting runtest to False
        if self.cfg['runtest']:
            if self.toolchain.options.get('usempi', None):
                cmd = "cd test && ./test.com M %s %s" % (self.cfg['parallel'], self.arch)
            else:
                cmd = "cd test && ./test.com %s" % self.arch
            (out, _) = run_cmd(cmd, log_all=True, simple=False)
            return out

    def sanity_check_step(self):
        """Custom sanity check for CHARMM."""
        custom_paths = {
            'files': [],
            'dirs': [],
        }
        if self.toolchain.options.get('usempi', None):
            custom_paths['files'].append('exec/%s_M/charmm' % self.arch)
        else:
            custom_paths['files'].append('exec/%s/charmm' % self.arch)
        super(EB_CHARMM, self).sanity_check_step(custom_paths=custom_paths)

    def install_step(self):
        """Copy the build directory to the install path"""
        self.log.info("Copying CHARMM dir %s to %s" % (self.cfg['start_dir'], self.installdir))
        try:
            shutil.copytree(self.cfg['start_dir'], self.installdir)
        except OSError as err:
            raise EasyBuildError("Failed to copy CHARMM dir to install dir: %s", err)

    def make_module_req_guess(self):
        """Custom guesses for environment variable PATH for CHARMM."""
        guesses = super(EB_CHARMM, self).make_module_req_guess()
        if self.toolchain.options.get('usempi', None):
            suffix = "_M"
        else:
            suffix = ""
        guesses.update({
            'PATH': ['exec/%s%s' % (self.arch, suffix)],
        })
        return guesses



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for OpenSSL, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Davide Vanzo (ACCRE - Vanderbilt University)
"""
import os
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd


class EB_OpenSSL(ConfigureMake):
    """Support for building OpenSSL"""

    def configure_step(self, cmd_prefix=''):
        """
        Configure step
        """

        cmd = "%s %s./config --prefix=%s threads shared %s" % (self.cfg['preconfigopts'], cmd_prefix,
                                                               self.installdir, self.cfg['configopts'])

        (out, _) = run_cmd(cmd, log_all=True, simple=False)

        return out

    def sanity_check_step(self):
        """Custom sanity check"""

        libdir = None
        for libdir_cand in ['lib', 'lib64']:
            if os.path.exists(os.path.join(self.installdir, libdir_cand)):
                libdir = libdir_cand

        if libdir is None:
            raise EasyBuildError("Failed to determine library directory.")

        custom_paths = {
            'files': [os.path.join(libdir, x) for x in ['libcrypto.a', 'libcrypto.so', 'libssl.a', 'libssl.so']] +
            ['bin/openssl'],
            'dirs': [],
        }

        if LooseVersion(self.version) < LooseVersion("1.1"):
            custom_paths['files'].extend([os.path.join(libdir, 'libcrypto.so.1.0.0'),
                                          os.path.join(libdir, 'libssl.so.1.0.0')])
            custom_paths['dirs'].append(os.path.join(libdir, 'engines'))
        else:
            custom_paths['files'].extend([os.path.join(libdir, 'libcrypto.so.1.1'),
                                          os.path.join(libdir, 'libssl.so.1.1')])
            custom_paths['dirs'].append(os.path.join(libdir, 'engines-1.1'))

        super(EB_OpenSSL, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing ESMF, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Damian Alvarez (Forschungszentrum Juelich GmbH)
"""
import os
from distutils.version import LooseVersion

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_ESMF(ConfigureMake):
    """Support for building/installing ESMF."""

    def configure_step(self):
        """Custom configuration procedure for ESMF through environment variables."""

        env.setvar('ESMF_DIR', self.cfg['start_dir'])
        env.setvar('ESMF_INSTALL_PREFIX', self.installdir)
        env.setvar('ESMF_INSTALL_BINDIR', 'bin')
        env.setvar('ESMF_INSTALL_LIBDIR', 'lib')
        env.setvar('ESMF_INSTALL_MODDIR', 'mod')

        # specify compiler
        comp_family = self.toolchain.comp_family()
        if comp_family in [toolchain.GCC]:
            compiler = 'gfortran'
        else:
            compiler = comp_family.lower()
        env.setvar('ESMF_COMPILER', compiler)

        env.setvar('ESMF_F90COMPILEOPTS', os.getenv('F90FLAGS'))
        env.setvar('ESMF_CXXCOMPILEOPTS', os.getenv('CXXFLAGS'))

        # specify MPI communications library
        comm = None
        mpi_family = self.toolchain.mpi_family()
        if mpi_family in [toolchain.MPICH, toolchain.QLOGICMPI]:
            # MPICH family for MPICH v3.x, which is MPICH2 compatible
            comm = 'mpich2'
        else:
            comm = mpi_family.lower()
        env.setvar('ESMF_COMM', comm)

        # specify decent LAPACK lib
        env.setvar('ESMF_LAPACK', 'user')
        ldflags = os.getenv('LDFLAGS')
        liblapack = os.getenv('LIBLAPACK_MT') or os.getenv('LIBLAPACK')
        if liblapack is None:
            raise EasyBuildError("$LIBLAPACK(_MT) not defined, no BLAS/LAPACK in %s toolchain?", self.toolchain.name)
        else:
            env.setvar('ESMF_LAPACK_LIBS', ldflags + ' ' + liblapack)

        # specify netCDF
        netcdf = get_software_root('netCDF')
        if netcdf:
            if LooseVersion(self.version) >= LooseVersion('7.1.0'):
                env.setvar('ESMF_NETCDF', 'nc-config')
            else:
                env.setvar('ESMF_NETCDF', 'user')
                netcdf_libs = ['-L%s/lib' % netcdf, '-lnetcdf']

                # Fortran
                netcdff = get_software_root('netCDF-Fortran')
                if netcdff:
                    netcdf_libs = ["-L%s/lib" % netcdff] + netcdf_libs + ["-lnetcdff"]
                else:
                    netcdf_libs.append('-lnetcdff')

                # C++
                netcdfcxx = get_software_root('netCDF-C++')
                if netcdfcxx:
                    netcdf_libs = ["-L%s/lib" % netcdfcxx] + netcdf_libs + ["-lnetcdf_c++"]
                else:
                    netcdfcxx = get_software_root('netCDF-C++4')
                    if netcdfcxx:
                        netcdf_libs = ["-L%s/lib" % netcdfcxx] + netcdf_libs + ["-lnetcdf_c++4"]
                    else:
                        netcdf_libs.append('-lnetcdf_c++')
                env.setvar('ESMF_NETCDF_LIBS', ' '.join(netcdf_libs))

        # 'make info' provides useful debug info
        cmd = "make info"
        run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def sanity_check_step(self):
        """Custom sanity check for ESMF."""

        binaries = ['ESMF_Info', 'ESMF_InfoC', 'ESMF_RegridWeightGen', 'ESMF_WebServController']
        libs = ['libesmf.a', 'libesmf.%s' % get_shared_lib_ext()]
        custom_paths = {
            'files': [os.path.join('bin', x) for x in binaries] + [os.path.join('lib', x) for x in libs],
            'dirs': ['include', 'mod'],
        }

        super(EB_ESMF, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing VEP, implemented as an easyblock
"""
import os

import easybuild.tools.environment as env
from easybuild.easyblocks.perl import get_major_perl_version
from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.filetools import apply_regex_substitutions
from easybuild.tools.modules import get_software_version
from easybuild.tools.run import run_cmd


class EB_VEP(EasyBlock):
    """Support for building/installing VEP."""

    def __init__(self, *args, **kwargs):
        """VEP easyblock constructor."""
        super(EB_VEP, self).__init__(*args, **kwargs)

        self.build_in_installdir = True
        self.cfg['unpack_options'] = "--strip-components=1"

        self.api_mods_subdir = os.path.join('modules', 'api')

    def configure_step(self):
        """No custom configuration procedure for VEP."""
        pass

    def build_step(self):
        """No custom build procedure for VEP."""
        pass

    def install_step(self):
        """Custom install procedure for VEP."""

        # patch INSTALL.pl script to use https:// rather than ftp://
        apply_regex_substitutions('INSTALL.pl', [('ftp://', 'https://')])

        # update PERL5LIB so tests can run (done automatically by INSTALL.pl unless --NO_TEST is used)
        perl_majver = get_major_perl_version()
        perllib_envvar = 'PERL%sLIB' % perl_majver
        perllib = os.getenv(perllib_envvar, '')
        api_mods_dir = os.path.join(self.installdir, self.api_mods_subdir)
        self.log.info("Adding %s to $%s (%s)", api_mods_dir, perllib_envvar, perllib)
        env.setvar(perllib_envvar, '%s:%s' % (api_mods_dir, perllib))

        # see https://www.ensembl.org/info/docs/tools/vep/script/vep_download.html#installer
        cmd = ' '.join([
            self.cfg['preinstallopts'],
            'perl',
            'INSTALL.pl',
            # don't try to install optional Bio::DB::HTS (can be provided as an extension instead)
            '--NO_HTSLIB',
            # a: API, f: FASTA
            # not included:
            # c: cache, should be downloaded by user
            # l: Bio::DB::HTS, should be provided via EasyBuild
            # p: plugins
            '--AUTO af',
            # install all species
            '--SPECIES all',
            # don't update VEP during installation
            '--NO_UPDATE',
            # location to install Perl API modules into
            '--DESTDIR ' + api_mods_dir,
            self.cfg['installopts'],
        ])
        run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def sanity_check_step(self):
        """Custom sanity check for VEP."""

        custom_paths = {
            'files': ['vep'],
            'dirs': ['modules/Bio/EnsEMBL/VEP'],
        }

        if 'Bio::EnsEMBL::XS' in [ext[0] for ext in self.cfg['exts_list']]:
            perl_majver = get_major_perl_version()
            perl_ver = get_software_version('Perl')
            perl_libpath = os.path.join('lib', 'perl' + perl_majver, 'site_perl', perl_ver)
            bio_ensembl_xs_ext = os.path.join(perl_libpath, 'x86_64-linux-thread-multi', 'Bio', 'EnsEMBL', 'XS.pm')
            custom_paths['files'].extend([bio_ensembl_xs_ext])

        custom_commands = ['vep --help']

        super(EB_VEP, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_req_guess(self):
        """Custom guesses for environment variables (PATH, ...) for VEP."""
        perl_majver = get_major_perl_version()

        perl_libpath = [self.api_mods_subdir]
        if 'Bio::EnsEMBL::XS' in [ext[0] for ext in self.cfg['exts_list']]:
            perl_ver = get_software_version('Perl')
            perl_libpath.extend([os.path.join('lib', 'perl' + perl_majver, 'site_perl', perl_ver)])

        guesses = super(EB_VEP, self).make_module_req_guess()
        guesses = {
            'PATH': '',
            'PERL%sLIB' % perl_majver: perl_libpath,
        }
        return guesses



##
# Copyright 2013-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for DL_POLY Classic, implemented as an easyblock

@author: Jens Timmerman (Ghent University)
@author: Kenneth Hoste (Ghent University)
"""
import glob
import os

from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import copy_file, copy_dir
from easybuild.tools.run import run_cmd
from easybuild.easyblocks.generic.configuremake import ConfigureMake


class EB_DL_underscore_POLY_underscore_Classic(ConfigureMake):
    """Support for building and installing DL_POLY Classic."""

    def __init__(self, *args, **kwargs):
        """Easyblock constructor; initialize class variables."""
        super(EB_DL_underscore_POLY_underscore_Classic, self).__init__(*args, **kwargs)

        # check whether PLUMED is listed as a dependency
        self.with_plumed = 'PLUMED' in [dep['name'] for dep in self.cfg['dependencies']]

    # create PLUMED patch in prepare_step rather than patch_step,
    # so we can rely on being in the unpacked source directory
    def prepare_step(self, *args, **kwargs):
        """Generate PLUMED patch if PLUMED is listed as a dependency."""
        super(EB_DL_underscore_POLY_underscore_Classic, self).prepare_step(*args, **kwargs)

        if self.with_plumed:
            # see https://groups.google.com/d/msg/plumed-users/cWaIDU5F6Bw/bZUW3J9cCAAJ
            diff_pat = 'dlpoly-*.diff'
            try:
                diff_hits = glob.glob(os.path.join(self.builddir, diff_pat))
            except OSError as err:
                raise EasyBuildError("Failed to find list of files/dirs that match '%s': %s", diff_pat, err)

            if len(diff_hits) == 1:
                plumed_patch = diff_hits[0]
            elif not self.dry_run:
                raise EasyBuildError("Expected to find exactly one match for '%s' in %s, found: %s",
                                     diff_pat, self.builddir, diff_hits)

            if not self.dry_run:
                try:
                    os.rename('source', 'srcmod')
                except OSError as err:
                    raise EasyBuildError("Failed to move 'source' directory to 'srcmod': %s", err)

            engine = os.path.splitext(os.path.basename(plumed_patch))[0]
            cmd = "plumed-patch -p --runtime -e %s -d %s" % (engine, plumed_patch)
            run_cmd(cmd, log_all=True, simple=True)

    def configure_step(self):
        """Copy the makefile to the source directory and use MPIF90 to do a parrallel build"""

        self.cfg.update('buildopts', 'LD="$MPIF90 -o" FC="$MPIF90 -c" par')

        if self.with_plumed:
            source_dir = 'srcmod'
            self.cfg.update('buildopts', 'LDFLAGS="${LDFLAGS} -lplumed -ldl"')
        else:
            source_dir = 'source'

        copy_file(os.path.join('build', 'MakePAR'), os.path.join(source_dir, 'Makefile'))
        try:
            os.chdir(source_dir)
        except OSError as err:
            raise EasyBuildError("Failed to change to %s: %s", source_dir, err)

    def install_step(self):
        """Copy the executables to the installation directory"""
        self.log.debug("copying %s/execute to %s, (from %s)", self.cfg['start_dir'], self.installdir, os.getcwd())
        # create a 'bin' subdir, this way we also get $PATH to be set correctly automatically
        install_path = os.path.join(self.cfg['start_dir'], 'execute')
        bin_path = os.path.join(self.installdir, 'bin')
        copy_dir(install_path, bin_path)

    def sanity_check_step(self):
        """Custom sanity check step for DL_POLY Classic"""
        custom_paths = {
            'files': ['bin/DLPOLY.X'],
            'dirs': [],
        }
        super(EB_DL_underscore_POLY_underscore_Classic, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for software that is configured with CMake, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
@author: Maxime Boissonneault (Compute Canada - Universite Laval)
"""
import os

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import print_warning
from easybuild.tools.config import build_option
from easybuild.tools.filetools import change_dir, create_unused_dir, mkdir, which
from easybuild.tools.environment import setvar
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext
from easybuild.tools.utilities import nub


DEFAULT_CONFIGURE_CMD = 'cmake'


def setup_cmake_env(tc):
    """Setup env variables that cmake needs in an EasyBuild context."""

    # Set the search paths for CMake
    tc_ipaths = tc.get_variable("CPPFLAGS", list)
    tc_lpaths = tc.get_variable("LDFLAGS", list)
    cpaths = os.getenv('CPATH', '').split(os.pathsep)
    lpaths = os.getenv('LD_LIBRARY_PATH', '').split(os.pathsep)
    include_paths = os.pathsep.join(nub(tc_ipaths + cpaths))
    library_paths = os.pathsep.join(nub(tc_lpaths + lpaths))
    setvar("CMAKE_INCLUDE_PATH", include_paths)
    setvar("CMAKE_LIBRARY_PATH", library_paths)


class CMakeMake(ConfigureMake):
    """Support for configuring build with CMake instead of traditional configure script"""

    @staticmethod
    def extra_options(extra_vars=None):
        """Define extra easyconfig parameters specific to CMakeMake."""
        extra_vars = ConfigureMake.extra_options(extra_vars)
        extra_vars.update({
            'abs_path_compilers': [False, "Specify compilers via absolute file path (not via command names)", CUSTOM],
            'allow_system_boost': [False, "Always allow CMake to pick up on Boost installed in OS "
                                          "(even if Boost is included as a dependency)", CUSTOM],
            'build_shared_libs': [None, "Build shared library (instead of static library)"
                                        "None can be used to add no flag (usually results in static library)", CUSTOM],
            'build_type': [None, "Build type for CMake, e.g. Release."
                                 "Defaults to 'Release' or 'Debug' depending on toolchainopts[debug]", CUSTOM],
            'configure_cmd': [DEFAULT_CONFIGURE_CMD, "Configure command to use", CUSTOM],
            'generator': [None, "Build file generator to use. None to use CMakes default", CUSTOM],
            'srcdir': [None, "Source directory location to provide to cmake command", CUSTOM],
            'separate_build_dir': [True, "Perform build in a separate directory", CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Constructor for CMakeMake easyblock"""
        super(CMakeMake, self).__init__(*args, **kwargs)
        self._lib_ext = None

    @property
    def lib_ext(self):
        """Return the extension for libraries build based on `build_shared_libs` or None if that is unset"""
        if self._lib_ext is None:
            build_shared_libs = self.cfg.get('build_shared_libs')
            if build_shared_libs:
                self._lib_ext = get_shared_lib_ext()
            elif build_shared_libs is not None:
                self._lib_ext = 'a'
        return self._lib_ext

    @lib_ext.setter
    def lib_ext(self, value):
        self._lib_ext = value

    @property
    def build_type(self):
        """Build type set in the EasyConfig with default determined by toolchainopts"""
        build_type = self.cfg.get('build_type')
        if build_type is None:
            build_type = 'Debug' if self.toolchain.options.get('debug', None) else 'Release'
        return build_type

    def configure_step(self, srcdir=None, builddir=None):
        """Configure build using cmake"""

        setup_cmake_env(self.toolchain)

        if builddir is None and self.cfg.get('separate_build_dir', True):
            builddir = create_unused_dir(self.builddir, 'easybuild_obj')

        if builddir:
            mkdir(builddir, parents=True)
            change_dir(builddir)
            default_srcdir = self.cfg['start_dir']
        else:
            default_srcdir = '.'

        if srcdir is None:
            if self.cfg.get('srcdir', None) is not None:
                # Note that the join returns srcdir if it is absolute
                srcdir = os.path.join(default_srcdir, self.cfg['srcdir'])
            else:
                srcdir = default_srcdir

        options = ['-DCMAKE_INSTALL_PREFIX=%s' % self.installdir]

        if self.installdir.startswith('/opt') or self.installdir.startswith('/usr'):
            # https://cmake.org/cmake/help/latest/module/GNUInstallDirs.html
            localstatedir = os.path.join(self.installdir, 'var')
            runstatedir = os.path.join(localstatedir, 'run')
            sysconfdir = os.path.join(self.installdir, 'etc')
            options.append("-DCMAKE_INSTALL_LOCALSTATEDIR=%s" % localstatedir)
            options.append("-DCMAKE_INSTALL_RUNSTATEDIR=%s" % runstatedir)
            options.append("-DCMAKE_INSTALL_SYSCONFDIR=%s" % sysconfdir)

        if '-DCMAKE_BUILD_TYPE=' in self.cfg['configopts']:
            if self.cfg.get('build_type') is not None:
                self.log.warning('CMAKE_BUILD_TYPE is set in configopts. Ignoring build_type')
        else:
            options.append('-DCMAKE_BUILD_TYPE=%s' % self.build_type)

        # Add -fPIC flag if necessary
        if self.toolchain.options['pic']:
            options.append('-DCMAKE_POSITION_INDEPENDENT_CODE=ON')

        if self.cfg['generator']:
            options.append('-G "%s"' % self.cfg['generator'])

        # pass --sysroot value down to CMake,
        # and enable using absolute paths to compiler commands to avoid
        # that CMake picks up compiler from sysroot rather than toolchain compiler...
        sysroot = build_option('sysroot')
        if sysroot:
            options.append('-DCMAKE_SYSROOT=%s' % sysroot)
            self.log.info("Using absolute path to compiler commands because of alterate sysroot %s", sysroot)
            self.cfg['abs_path_compilers'] = True

        # Set flag for shared libs if requested
        # Not adding one allows the project to choose a default
        build_shared_libs = self.cfg.get('build_shared_libs')
        if build_shared_libs is not None:
            # Contrary to other options build_shared_libs takes precedence over configopts which may be unexpected.
            # This is to allow self.lib_ext to be determined correctly.
            # Usually you want to remove -DBUILD_SHARED_LIBS from configopts and set build_shared_libs to True or False
            # If you need it in configopts don't set build_shared_libs (or explicitely set it to `None` (Default))
            if '-DBUILD_SHARED_LIBS=' in self.cfg['configopts']:
                print_warning('Ignoring BUILD_SHARED_LIBS is set in configopts because build_shared_libs is set')
            self.cfg.update('configopts', '-DBUILD_SHARED_LIBS=%s' % ('ON' if build_shared_libs else 'OFF'))

        env_to_options = {
            'CC': 'CMAKE_C_COMPILER',
            'CFLAGS': 'CMAKE_C_FLAGS',
            'CXX': 'CMAKE_CXX_COMPILER',
            'CXXFLAGS': 'CMAKE_CXX_FLAGS',
            'F90': 'CMAKE_Fortran_COMPILER',
            'FFLAGS': 'CMAKE_Fortran_FLAGS',
        }
        for env_name, option in env_to_options.items():
            value = os.getenv(env_name)
            if value is not None:
                if option.endswith('_COMPILER') and self.cfg.get('abs_path_compilers', False):
                    value = which(value)
                    self.log.info("Using absolute path to compiler command: %s", value)
                options.append("-D%s='%s'" % (option, value))

        if build_option('rpath'):
            # instruct CMake not to fiddle with RPATH when --rpath is used, since it will undo stuff on install...
            # https://github.com/LLNL/spack/blob/0f6a5cd38538e8969d11bd2167f11060b1f53b43/lib/spack/spack/build_environment.py#L416
            options.append('-DCMAKE_SKIP_RPATH=ON')

        # show what CMake is doing by default
        options.append('-DCMAKE_VERBOSE_MAKEFILE=ON')

        # disable CMake user package repository
        options.append('-DCMAKE_FIND_USE_PACKAGE_REGISTRY=FALSE')

        if not self.cfg.get('allow_system_boost', False):
            # don't pick up on system Boost if Boost is included as dependency
            # - specify Boost location via -DBOOST_ROOT
            # - instruct CMake to not search for Boost headers/libraries in other places
            # - disable search for Boost CMake package configuration file
            boost_root = get_software_root('Boost')
            if boost_root:
                options.extend([
                    '-DBOOST_ROOT=%s' % boost_root,
                    '-DBoost_NO_SYSTEM_PATHS=ON',
                    '-DBoost_NO_BOOST_CMAKE=ON',
                ])

        options_string = ' '.join(options)

        if self.cfg.get('configure_cmd') == DEFAULT_CONFIGURE_CMD:
            command = ' '.join([
                self.cfg['preconfigopts'],
                DEFAULT_CONFIGURE_CMD,
                options_string,
                self.cfg['configopts'],
                srcdir])
        else:
            command = ' '.join([
                self.cfg['preconfigopts'],
                self.cfg.get('configure_cmd'),
                self.cfg['configopts']])

        (out, _) = run_cmd(command, log_all=True, simple=False)

        return out

    def test_step(self):
        """CMake specific test setup"""
        # When using ctest for tests (default) then show verbose output if a test fails
        setvar('CTEST_OUTPUT_ON_FAILURE', 'True')
        super(CMakeMake, self).test_step()



##
# Copyright 2015-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Cray toolchains, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Guilherme Peretti Pezzi (CSCS)
@author: Petar Forai (IMP/IMBA)
"""

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.tools.build_log import EasyBuildError


KNOWN_PRGENVS = ['PrgEnv-cray', 'PrgEnv-gnu', 'PrgEnv-intel', 'PrgEnv-pgi']


class CrayToolchain(Bundle):
    """
    Compiler toolchain: generate module file only, nothing to build/install
    """

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment (skip loaded of dependencies)."""

        kwargs['load_tc_deps_modules'] = False

        super(CrayToolchain, self).prepare_step(*args, **kwargs)

    def make_module_dep(self):
        """
        Generate load/swap statements for dependencies in the module file
        """
        prgenv_mod = None

        # collect 'swap' statement for dependencies (except PrgEnv)
        swap_deps = []
        for dep in self.toolchain.dependencies:
            mod_name = dep['full_mod_name']
            # determine versionless module name, e.g. 'fftw/3.3.4.1' => 'fftw'
            dep_name = '/'.join(mod_name.split('/')[:-1])

            if mod_name.startswith('PrgEnv'):
                prgenv_mod = mod_name
            else:
                swap_deps.append(self.module_generator.swap_module(dep_name, mod_name).lstrip())

        self.log.debug("Swap statements for dependencies of %s: %s", self.full_mod_name, swap_deps)

        if prgenv_mod is None:
            raise EasyBuildError("Could not find a PrgEnv-* module listed as dependency: %s",
                                 self.toolchain.dependencies)

        # unload statements for other PrgEnv modules
        prgenv_unloads = ['']
        for prgenv in [prgenv for prgenv in KNOWN_PRGENVS if not prgenv_mod.startswith(prgenv)]:
            is_loaded_guard = self.module_generator.is_loaded(prgenv)
            unload_stmt = self.module_generator.unload_module(prgenv).strip()
            prgenv_unloads.append(self.module_generator.conditional_statement(is_loaded_guard, unload_stmt))

        # load statement for selected PrgEnv module (only when not loaded yet)
        prgenv_load = self.module_generator.load_module(prgenv_mod, recursive_unload=False)

        txt = '\n'.join(prgenv_unloads + [prgenv_load] + swap_deps)
        return txt



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing deepmind/reverb, implemented as an easyblock
@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os

from easybuild.easyblocks.generic.pythonpackage import PythonPackage, PIP_INSTALL_CMD
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.environment import setvar
from easybuild.tools.filetools import mkdir, remove_dir
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd


class EB_dm_minus_reverb(PythonPackage):
    """Support for installing deepmind/reverb Python package."""

    def configure_step(self, *args, **kwargs):
        """Execute custom configure.py script"""
        super(EB_dm_minus_reverb, self).configure_step(*args, **kwargs)

        # set Python environment
        python_root = get_software_root('Python')
        if python_root:
            pyshortver = '.'.join(get_software_version('Python').split('.')[:2])
            pylibdir = os.path.join(self.installdir, 'lib', 'python%s' % pyshortver, 'site-packages')
            setvar('PYTHON_LIB_PATH', pylibdir)
            pybin = os.path.join(python_root, 'bin', 'python')
            setvar('PYTHON_BIN_PATH', pybin)
        else:
            raise EasyBuildError("Python not found in dependency list.")

        # use protobuf version from EB
        if get_software_root('protobuf'):
            setvar('REVERB_PROTOC_VERSION', get_software_version('protobuf'))

        # execute custom configuration script
        conf_cmd = "python configure.py"
        return run_cmd(conf_cmd, log_all=True, simple=True, log_output=True)

    def build_step(self, *args, **kwargs):
        """Build with Bazel"""
        if not get_software_root('Bazel'):
            raise EasyBuildError("Bazel not found in dependency list.")

        # separate build dir for Bazel
        build_dir = os.path.join(self.builddir, 'easybuild_obj')
        if os.path.exists(build_dir):
            self.log.warning('Build directory %s already exists (from previous iterations?). Removing...', build_dir)
            remove_dir(build_dir)
        mkdir(build_dir, parents=True)
        bazel_opts = "--output_user_root %s" % build_dir

        # build target
        bazel_build_pkg = '//reverb/pip_package:build_pip_package'

        # generate build command
        bazel_build_opts = self.cfg['buildopts']
        # by default generate a release build
        if not all(opt in bazel_build_opts for opt in (" --compilation_mode", " -c")):
            bazel_build_opts += " --compilation_mode=opt"
        # set C++ standard (--cxxopt can be used multiple times)
        cstd = self.toolchain.options.get('cstd', None)
        if cstd:
            bazel_build_opts += " --cxxopt='-std=%s'" % cstd
        # use JDK from EB
        bazel_build_opts += " --host_javabase=@local_jdk//:jdk"
        # explicitly set the number of processes
        bazel_build_opts += " --jobs=%d" % self.cfg['parallel']
        # print full compilation commands
        bazel_build_opts += " --subcommands"

        bazel_cmd = "bazel %s build %s %s" % (bazel_opts, bazel_build_opts, bazel_build_pkg)

        return run_cmd(bazel_cmd, log_all=True, simple=True, log_output=True)

    def install_step(self, *args, **kwargs):
        """Package deepmind/reverb in a wheel and install it with pip"""

        # package a release version
        whl_build_opts = "--release"
        whl_build_opts += " --dst %s/" % self.builddir
        # target TF from EasyBuild
        if get_software_root('TensorFlow'):
            whl_build_opts += " --tf-version %s" % get_software_version('TensorFlow')

        whl_cmd = "./bazel-bin/reverb/pip_package/build_pip_package %s" % whl_build_opts

        run_cmd(whl_cmd, log_all=True, simple=True, log_output=True)

        # install wheel with pip
        pymajmin = ''.join(get_software_version('Python').split('.')[:2])
        whl_file = '%s-%s-cp%s*.whl' % (self.name.replace('-', '_'), self.version, pymajmin)
        whl_path = os.path.join(self.builddir, whl_file)

        self.install_cmd = PIP_INSTALL_CMD % {
            'installopts': self.cfg['installopts'],
            'loc': whl_path,
            'prefix': self.installdir,
        }

        return super(EB_dm_minus_reverb, self).install_step(*args, **kwargs)



##
# Copyright 2009-2021 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing RPMs, implemented as an easyblock.

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Toon Willems (Ghent University)
"""

import glob
import os
import re
import tempfile
from distutils.version import LooseVersion
from os.path import expanduser

import easybuild.tools.environment as env
from easybuild.base import fancylogger
from easybuild.easyblocks.generic.binary import Binary
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import change_dir, mkdir, symlink, which
from easybuild.tools.run import run_cmd


_log = fancylogger.getLogger('easyblocks.generic.rpm')


def rebuild_rpm(rpm_path, targetdir):
    """Rebuild the RPM on the specified location, to make it relocatable."""
    # make sure that rpmrebuild command is available
    if not which('rpmrebuild'):
        raise EasyBuildError("Command 'rpmrebuild' is required but not available. " +
                             "Please add it as a dependency or install it with the OS package manager.")

    rpmmacros = os.path.join(expanduser('~'), '.rpmmacros')
    if os.path.exists(rpmmacros):
        raise EasyBuildError("rpmmacros file %s found which will override any other settings, so exiting.", rpmmacros)

    rpmrebuild_tmpdir = os.path.join(tempfile.gettempdir(), "rpmrebuild")
    env.setvar("RPMREBUILD_TMPDIR", rpmrebuild_tmpdir)

    try:
        if not os.path.exists(rpmrebuild_tmpdir):
            os.makedirs(rpmrebuild_tmpdir)
            _log.debug("Created RPMREBUILD_TMPDIR dir %s" % rpmrebuild_tmpdir)
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)
            _log.debug("Created target directory for rebuilt RPMs %s" % targetdir)
    except OSError as err:
        raise EasyBuildError("Failed to create directories for rebuilding RPM: %s", err)

    _log.debug("Rebuilding %s in %s to make it relocatable" % (rpm_path, targetdir))
    cmd = ' '.join([
        "rpmrebuild -v",
        # replace whathever prefix is set with '/'
        r"""--change-spec-whole='sed -e "s/^Prefix:.*/Prefix: \//"'""",
        # comment out any specifications that involve relative file path (starting with '.') (??)
        r"""--change-spec-whole='sed -e "s/^\(.*:[ ]\+\..*\)/#ERROR \1/"'""",
        "--notest-install",
        "-p -d",
        targetdir,
        rpm_path,
    ])
    run_cmd(cmd, log_all=True, simple=True)


class Rpm(Binary):
    """
    Support for installing RPM files.
    - sources is a list of rpms
    - installation is with --nodeps (so the sources list has to be complete)
    """

    def __init__(self, *args, **kwargs):
        """Initialize class variables."""
        super(Rpm, self).__init__(*args, **kwargs)

        self.rebuild_rpm = False

    @staticmethod
    def extra_options(extra_vars=None):
        """Extra easyconfig parameters specific to RPMs."""
        extra_vars = Binary.extra_options(extra_vars)
        extra_vars.update({
            'force': [False, "Use force", CUSTOM],
            'preinstall': [False, "Enable pre install", CUSTOM],
            'postinstall': [False, "Enable post install", CUSTOM],
            'makesymlinks': [[], "Create symlinks for listed paths", CUSTOM],  # supports glob
        })
        return extra_vars

    def configure_step(self):
        """Custom configuration procedure for RPMs: rebuild RPMs for relocation if required."""

        # make sure that rpm is available
        if not which('rpm'):
            raise EasyBuildError("Command 'rpm' is required but not available.")

        # determine whether RPMs need to be rebuilt to make relocation work
        cmd = "rpm --version"
        (out, _) = run_cmd(cmd, log_all=True, simple=False)

        rpmver_re = re.compile(r"^RPM\s+version\s+(?P<version>[0-9.]+).*")
        res = rpmver_re.match(out)
        self.log.debug("RPM version found: %s" % res.group())

        if res:
            ver = res.groupdict()['version']

            # rebuilding is required on SL6, which implies rpm v4.8 (works fine without rebuilding on SL5)
            if LooseVersion(ver) >= LooseVersion('4.8.0'):
                self.rebuild_rpm = True
                self.log.debug("Enabling rebuild of RPMs to make relocation work...")
        else:
            raise EasyBuildError("Checking RPM version failed, so just carrying on with the default behaviour...")

        if self.rebuild_rpm:
            self.rebuild_rpms()

    # when installing RPMs under a non-default path for e.g. SL6,
    # --relocate doesn't seem to work (error: Unable to change root directory: Operation not permitted)
    def rebuild_rpms(self):
        """Rebuild RPMs to make relocation work."""
        for rpm in self.src:
            rebuild_rpm(rpm['path'], targetdir=self.builddir)

        self.oldsrc = self.src
        self.src = []
        for path in glob.glob(os.path.join(self.builddir, '*', '*.rpm')):
            self.src.append({
                'name': os.path.basename(path),
                'path': path,
            })
        self.log.debug("oldsrc: %s, src: %s" % (self.oldsrc, self.src))

    def install_step(self):
        """Custom installation procedure for RPMs into a custom prefix."""

        change_dir(self.installdir)
        mkdir('rpm')

        cmd = "rpm --initdb --dbpath /rpm --root %s" % self.installdir

        run_cmd(cmd, log_all=True, simple=True)

        force = ''
        if self.cfg['force']:
            force = '--force'

        postinstall = '--nopost'
        if self.cfg['postinstall']:
            postinstall = ''
        preinstall = '--nopre'
        if self.cfg['preinstall']:
            preinstall = ''

        if self.rebuild_rpm:
            cmd_tpl = "rpm -i --dbpath %(inst)s/rpm %(force)s --relocate /=%(inst)s " \
                      "%(pre)s %(post)s --nodeps --ignorearch %(rpm)s"
        else:
            cmd_tpl = "rpm -i --dbpath /rpm %(force)s --root %(inst)s --relocate /=%(inst)s " \
                      "%(pre)s %(post)s --nodeps %(rpm)s"

        # exception for user root:
        # --relocate is not necessary -> --root will relocate more than enough
        # cmd_tpl = "rpm -i --dbpath /rpm %(force)s --root %(inst)s %(pre)s %(post)s --nodeps %(rpm)s"

        for rpm in self.src:
            cmd = cmd_tpl % {
                'inst': self.installdir,
                'rpm': rpm['path'],
                'force': force,
                'pre': preinstall,
                'post': postinstall,
            }
            run_cmd(cmd, log_all=True, simple=True)

        for path in self.cfg['makesymlinks']:
            # allow globs, always use first hit.
            # also verify links existince
            realdirs = glob.glob(path)
            if realdirs:
                if len(realdirs) > 1:
                    self.log.debug("More then one match found for symlink glob %s, using first (all: %s)",
                                   path, realdirs)
                symlink(realdirs[0], os.path.join(self.installdir, os.path.basename(path)))
            else:
                self.log.debug("No match found for symlink glob %s." % path)

    def make_module_req_guess(self):
        """Add common PATH/LD_LIBRARY_PATH paths found in RPMs to list of guesses."""

        guesses = super(Rpm, self).make_module_req_guess()

        guesses.update({
            'PATH': guesses.get('PATH', []) + ['usr/bin', 'sbin', 'usr/sbin'],
            'LD_LIBRARY_PATH': guesses.get('LD_LIBRARY_PATH', []) + ['usr/lib', 'usr/lib64'],
            'MANPATH': guesses.get('MANPATH', []) + ['usr/share/man'],
        })

        return guesses
